import dataclasses
import json
import os
import re
import traceback
from collections.abc import Sequence
from enum import StrEnum
from os import path
from typing import Any, TypeVar

import requests
from deepdiff.diff import DeepDiff
from dodal.utils import get_beamline_name
from jsonschema import ValidationError
from pydantic import BaseModel

from mx_bluesky.common.parameters.components import (
    WithVisit,
)
from mx_bluesky.common.parameters.constants import (
    GridscanParamConstants,
)
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.common.utils.utils import convert_angstrom_to_ev
from mx_bluesky.hyperion._plan_runner_params import Wait
from mx_bluesky.hyperion.blueapi.parameters import LoadCentreCollectParams
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect

T = TypeVar("T", bound=WithVisit)
MULTIPIN_PREFIX = "multipin"
MULTIPIN_FORMAT_DESC = "Expected multipin format is multipin_{number_of_wells}x{well_size}+{distance_between_tip_and_first_well}"
MULTIPIN_REGEX = rf"^{MULTIPIN_PREFIX}_(\d+)x(\d+(?:\.\d+)?)\+(\d+(?:\.\d+)?)$"
MX_GENERAL_ROOT_REGEX = r"^/dls/(?P<beamline>[^/]+)/data/[^/]*/(?P<visit>[^/]+)(?:/|$)"


class _InstructionType(StrEnum):
    WAIT = "wait"
    COLLECT = "collect"


@dataclasses.dataclass
class _PinType:
    expected_number_of_crystals: int
    single_well_width_um: float
    tip_to_first_well_um: float = 0

    @property
    def full_width(self) -> float:
        """This is the "width" of the area where there may be samples.

        From a pin perspective this is along the length of the pin but we use width here as
        we mount the sample at 90 deg to the optical camera.

        We calculate the full width by adding all the gaps between wells then assuming
        there is a buffer of {tip_to_first_well_um} either side too. In reality the
        calculation does not need to be very exact as long as we get a width that's good
        enough to use for optical centring and XRC grid size.
        """
        return (self.expected_number_of_crystals - 1) * self.single_well_width_um + (
            2 * self.tip_to_first_well_um
        )


class _SinglePin(_PinType):
    def __init__(self):
        super().__init__(1, GridscanParamConstants.WIDTH_UM)

    @property
    def full_width(self) -> float:
        return self.single_well_width_um


def create_parameters_from_agamemnon() -> Sequence[BaseModel]:
    """Fetch the next instruction from agamemnon and convert it into one or more
    mx-bluesky instructions.
    Returns:
        The generated sequence of mx-bluesky parameters, or empty list if
        no instructions."""
    beamline_name = get_beamline_name("i03")
    agamemnon_instruction = _get_next_instruction(beamline_name)
    if agamemnon_instruction:
        match _instruction_and_data(agamemnon_instruction):
            case (_InstructionType.COLLECT, data):
                return _populate_parameters_from_agamemnon(data)
            case (_InstructionType.WAIT, data):
                return [
                    Wait.model_validate(
                        {
                            "duration_s": data,
                        }
                    )
                ]

    return []


def compare_params(load_centre_collect_params: LoadCentreCollect):
    """Compare the supplied parameters (as supplied from GDA) with those directly
    created from agamemnon. Any differences are logged.
    Args:
        load_centre_collect_params: The parameters from GDA to compare."""
    try:
        lcc_requests = create_parameters_from_agamemnon()
        # Log differences against GDA populated parameters
        if not lcc_requests:
            LOGGER.info("Agamemnon returned no instructions")
        else:
            differences = DeepDiff(
                lcc_requests[0], load_centre_collect_params, math_epsilon=1e-5
            )
            if differences:
                LOGGER.info(
                    f"Different parameters found when directly reading from Hyperion: {differences}"
                )
    except (ValueError, KeyError):
        LOGGER.warning(f"Failed to compare parameters: {traceback.format_exc()}")
    except Exception:
        LOGGER.warning(
            f"Unexpected error occurred. Failed to compare parameters: {traceback.format_exc()}"
        )


def update_params_from_agamemnon(parameters: T) -> T:
    """Update the supplied parameters with additional information from agamemnon.
    This is currently necessary for multipin processing and called when Hyperion is invoked
    from GDA.

    Args:
        parameters: The LoadCentreCollectParameters that will be updated with additional info,
        such as multipin dimensions, number of crystals.
    """
    try:
        beamline_name = get_beamline_name("i03")
        agamemnon_params = _get_next_instruction(beamline_name)
        instruction, collect_params = _instruction_and_data(agamemnon_params)
        assert instruction == _InstructionType.COLLECT, (
            "Unable to augment GDA parameters from agamemnon, agamemnon reports 'wait'"
        )
        pin_type = _get_pin_type_from_agamemnon_collect_parameters(collect_params)
        if isinstance(parameters, LoadCentreCollect):
            parameters.robot_load_then_centre.tip_offset_um = pin_type.full_width / 2
            parameters.robot_load_then_centre.grid_width_um = pin_type.full_width
            parameters.select_centres.n = pin_type.expected_number_of_crystals
            if pin_type != _SinglePin():
                # Rotation snapshots will be generated from the gridscan snapshots,
                # no need to specify snapshot omega.
                parameters.multi_rotation_scan.snapshot_omegas_deg = []
                parameters.multi_rotation_scan.use_grid_snapshots = True
    except (ValueError, ValidationError) as e:
        LOGGER.warning(f"Failed to update parameters: {e}")
    except Exception as e:
        LOGGER.warning(f"Unexpected error occurred. Failed to update parameters: {e}")

    return parameters


def _instruction_and_data(agamemnon_instruction: dict) -> tuple[str, Any]:
    instruction, data = next(iter(agamemnon_instruction.items()))
    if instruction not in _InstructionType.__members__.values():
        raise KeyError(
            f"Unexpected instruction from agamemnon: {agamemnon_instruction}"
        )
    return instruction, data


def _get_parameters_from_url(url: str) -> dict:
    response = requests.get(url, headers={"Accept": "application/json"})
    response.raise_for_status()
    return json.loads(response.content)


def _get_pin_type_from_agamemnon_collect_parameters(
    collect_parameters: dict,
) -> _PinType:
    loop_type_name: str | None = collect_parameters["sample"]["loopType"]
    if loop_type_name:
        regex_search = re.search(MULTIPIN_REGEX, loop_type_name)
        if regex_search:
            wells, well_size, tip_to_first_well = regex_search.groups()
            return _PinType(int(wells), float(well_size), float(tip_to_first_well))
        else:
            loop_type_message = (
                f"Agamemnon loop type of {loop_type_name} not recognised"
            )
            if loop_type_name.startswith(MULTIPIN_PREFIX):
                raise ValueError(f"{loop_type_message}. {MULTIPIN_FORMAT_DESC}")
            LOGGER.warning(f"{loop_type_message}, assuming single pin")
    return _SinglePin()


def _get_next_instruction(beamline: str) -> dict:
    return _get_parameters_from_url(get_agamemnon_url() + f"getnextcollect/{beamline}")


def get_agamemnon_url() -> str:
    return os.environ.get("AGAMEMNON_URL", "http://agamemnon.diamond.ac.uk/")


def _get_withvisit_parameters_from_agamemnon(parameters: dict) -> tuple:
    try:
        prefix = parameters["prefix"]
        collection = parameters["collection"]
        # Assuming distance is identical for multiple collections. Remove after https://github.com/DiamondLightSource/mx-bluesky/issues/773
        detector_distance = collection[0]["distance"]
    except KeyError as e:
        raise KeyError("Unexpected json from agamemnon") from e

    match = re.match(MX_GENERAL_ROOT_REGEX, prefix) if prefix else None

    if match:
        return (match.group("visit"), detector_distance)

    raise ValueError(
        f"Agamemnon prefix '{prefix}' does not match MX-General root structure"
    )


def _get_withenergy_parameters_from_agamemnon(parameters: dict) -> dict[str, Any]:
    try:
        first_collection: dict = parameters["collection"][0]
        wavelength: float | None = first_collection.get("wavelength")
        assert isinstance(wavelength, float)
        demand_energy_ev = convert_angstrom_to_ev(wavelength)
        return {"demand_energy_ev": demand_energy_ev}
    except (KeyError, IndexError, AttributeError, TypeError):
        return {"demand_energy_ev": None}


def _populate_parameters_from_agamemnon(
    agamemnon_params,
) -> Sequence[LoadCentreCollectParams]:
    if not agamemnon_params:
        # Empty dict means no instructions
        return []

    visit, detector_distance = _get_withvisit_parameters_from_agamemnon(
        agamemnon_params
    )
    with_energy_params = _get_withenergy_parameters_from_agamemnon(agamemnon_params)
    pin_type = _get_pin_type_from_agamemnon_collect_parameters(agamemnon_params)
    collections = agamemnon_params["collection"]
    visit_directory, file_name = path.split(agamemnon_params["prefix"])

    return [
        LoadCentreCollectParams.model_validate(
            {
                "visit": visit,
                "detector_distance_mm": detector_distance,
                "sample_id": agamemnon_params["sample"]["id"],
                "sample_puck": agamemnon_params["sample"]["container"],
                "sample_pin": agamemnon_params["sample"]["position"],
                "select_centres": {
                    "name": "TopNByMaxCount",
                    "n": pin_type.expected_number_of_crystals,
                },
                **with_energy_params,
                "robot_load_then_centre": {
                    "storage_directory": str(visit_directory) + "/xraycentring",
                    "file_name": file_name,
                    "tip_offset_um": pin_type.full_width / 2,
                    "grid_width_um": pin_type.full_width,
                    "omega_start_deg": 0.0,
                    "chi_start_deg": collection["chi"],
                    "transmission_frac": 1.0,
                    "exposure_time_s": GridscanParamConstants.EXPOSURE_TIME_S,
                },
                "multi_rotation_scan": {
                    "comment": collection["comment"],
                    "storage_directory": str(visit_directory),
                    "exposure_time_s": collection["exposure_time"],
                    "file_name": file_name,
                    "transmission_frac": collection["transmission"],
                    "rotation_increment_deg": collection["omega_increment"],
                    "ispyb_experiment_type": collection["experiment_type"],
                    "snapshot_omegas_deg": [0.0, 90.0, 180.0, 270.0],
                    "rotation_scans": [
                        {
                            "scan_width_deg": (
                                collection["number_of_images"]
                                * collection["omega_increment"]
                            ),
                            "omega_start_deg": collection["omega_start"],
                            "phi_start_deg": collection["phi_start"],
                            "chi_start_deg": collection["chi"],
                            "rotation_direction": "Positive",
                        }
                    ],
                },
            }
        )
        for collection in collections
    ]



from unittest.mock import MagicMock, patch

import pytest

from mx_bluesky.beamlines.i04.callbacks.murko_callback import MurkoCallback


@pytest.fixture
def murko_callback() -> MurkoCallback:
    callback = MurkoCallback("", "")
    callback.redis_client = MagicMock()
    callback.redis_connected = True
    return callback


@pytest.fixture(autouse=True)
def always_use_i04_beamline(monkeypatch):
    monkeypatch.setenv("BEAMLINE", "i04")


@pytest.fixture(autouse=True)
def patch_get_i04_feature_settings():
    fake_path = "tests/test_data/test_domain_properties"
    with patch(
        "mx_bluesky.beamlines.i04.external_interaction.config_server.GDA_DOMAIN_PROPERTIES_PATH",
        str(fake_path),
    ):
        yield



from typing import Protocol, runtime_checkable

import bluesky.plan_stubs as bps
from bluesky.utils import MsgGenerator
from dodal.devices.zebra.zebra import (
    ArmDemand,
    EncEnum,
    I03Axes,
    RotationDirection,
    Zebra,
)
from dodal.devices.zebra.zebra_controlled_shutter import (
    ZebraShutter,
    ZebraShutterControl,
)

from mx_bluesky.common.parameters.constants import (
    ZEBRA_STATUS_TIMEOUT,
    PlanGroupCheckpointConstants,
)
from mx_bluesky.common.utils.log import LOGGER

"""Plans in this file will work as intended if the zebra has the following configuration:
- A fast shutter is connected through TTL inputs from the Zebra.
- When the zebra shutter is set to auto mode, the IOC sets the Zebra's SOFT_IN1 signal high.
- When the zebra shutter is set to manual mode, the IOC sets the Zebra's SOFT_IN1 signal low.
"""


@runtime_checkable
class GridscanSetupDevices(Protocol):
    zebra: Zebra
    sample_shutter: ZebraShutter


def setup_zebra_for_gridscan(
    composite: GridscanSetupDevices,  # XRC gridscan's generic trigger setup expects a composite rather than individual devices
    group="setup_zebra_for_gridscan",
    wait=True,
    ttl_input_for_detector_to_use: None | int = None,
) -> MsgGenerator:
    """
    Configure the zebra for an MX XRC gridscan by allowing the zebra to trigger the fast shutter and detector via signals
    sent from the motion controller.

    Args:
        composite: Composite device containing a zebra and zebra shutter
        group: Bluesky group to use when waiting on completion
        wait: If true, block until completion
        ttl_input_for_detector_to_use: If the zebra isn't using the TTL_DETECTOR zebra input, manually
        specify which TTL input is being used for the desired detector

    This plan assumes that the motion controller, as part of its gridscan PLC, will send triggers as required to the zebra's
    IN4_TTL and IN3_TTL to control the fast_shutter and detector respectively

    """
    zebra = composite.zebra
    ttl_detector = ttl_input_for_detector_to_use or zebra.mapping.outputs.TTL_DETECTOR
    # Set shutter to automatic and to trigger via motion controller GPIO signal (IN4_TTL)
    yield from configure_zebra_and_shutter_for_auto_shutter(
        zebra, composite.sample_shutter, zebra.mapping.sources.IN4_TTL, group=group
    )

    yield from bps.abs_set(
        zebra.output.out_pvs[ttl_detector],
        zebra.mapping.sources.IN3_TTL,
        group=group,
    )

    if wait:
        yield from bps.wait(group, timeout=ZEBRA_STATUS_TIMEOUT)


def set_shutter_auto_input(zebra: Zebra, input: int, group="set_shutter_trigger"):
    """Set the signal that controls the shutter. We use the second input to the
    Zebra's AND_GATE_FOR_AUTO_SHUTTER for this input. ZebraShutter control mode must be in auto for this input to take control

    For more details see the ZebraShutter device."""
    auto_gate = zebra.mapping.AND_GATE_FOR_AUTO_SHUTTER
    auto_shutter_control = zebra.logic_gates.and_gates[auto_gate]
    yield from bps.abs_set(auto_shutter_control.sources[2], input, group=group)


def configure_zebra_and_shutter_for_auto_shutter(
    zebra: Zebra, zebra_shutter: ZebraShutter, input: int, group="use_automatic_shutter"
):
    """Set the shutter to auto mode, and configure the zebra to trigger the shutter on
    an input source. For the input, use one of the source constants in zebra.py

    When the shutter is in auto/manual, logic in EPICS sets the Zebra's
    SOFT_IN1 to low/high respectively. The Zebra's AND_GATE_FOR_AUTO_SHUTTER should be used to control the shutter while in auto mode.
    To do this, we need (AND_GATE_FOR_AUTO_SHUTTER = SOFT_IN1 AND input), where input is the zebra signal we want to control the shutter when in auto mode.
    """

    # Set shutter to auto mode
    yield from bps.abs_set(
        zebra_shutter.control_mode, ZebraShutterControl.AUTO, group=group
    )

    auto_gate = zebra.mapping.AND_GATE_FOR_AUTO_SHUTTER

    # Set first input of AND_GATE_FOR_AUTO_SHUTTER to SOFT_IN1, which is high when shutter is in auto mode
    # Note the Zebra should ALWAYS be setup this way. See https://github.com/DiamondLightSource/mx-bluesky/issues/551
    yield from bps.abs_set(
        zebra.logic_gates.and_gates[auto_gate].sources[1],
        zebra.mapping.sources.SOFT_IN1,
        group=group,
    )

    # Set the second input of AND_GATE_FOR_AUTO_SHUTTER to the requested zebra input source
    yield from set_shutter_auto_input(zebra, input, group=group)


def tidy_up_zebra_after_gridscan(
    zebra: Zebra,
    zebra_shutter: ZebraShutter,
    group="tidy_up_zebra_after_gridscan",
    wait=True,
    ttl_input_for_detector_to_use: int | None = None,
) -> MsgGenerator:
    """
    Set the zebra back to a state which is expected by GDA.

    Args:
        zebra: Zebra device.
        zebra_shutter: Zebra shutter device.
        group: Bluesky group to use when waiting on completion.
        wait: If true, block until completion.
        ttl_input_for_detector_to_use: If the zebra isn't using the TTL_DETECTOR zebra input, manually
        specify which TTL input is being used for the desired detector.
    """

    LOGGER.info("Tidying up Zebra")

    ttl_detector = ttl_input_for_detector_to_use or zebra.mapping.outputs.TTL_DETECTOR

    yield from bps.abs_set(
        zebra.output.out_pvs[ttl_detector],
        zebra.mapping.sources.PC_PULSE,
        group=group,
    )
    yield from bps.abs_set(
        zebra_shutter.control_mode, ZebraShutterControl.MANUAL, group=group
    )
    yield from set_shutter_auto_input(zebra, zebra.mapping.sources.PC_GATE, group=group)

    if wait:
        yield from bps.wait(group, timeout=ZEBRA_STATUS_TIMEOUT)


def setup_zebra_for_rotation(
    zebra: Zebra,
    zebra_shutter: ZebraShutter,
    axis: EncEnum = I03Axes.OMEGA,
    start_angle: float = 0,
    scan_width: float = 360,
    shutter_opening_deg: float = 2.5,
    shutter_opening_s: float = 0.04,
    direction: RotationDirection = RotationDirection.POSITIVE,
    group: str = PlanGroupCheckpointConstants.SETUP_ZEBRA_FOR_ROTATION,
    wait: bool = True,
    ttl_input_for_detector_to_use: int | None = None,
):
    """Set up the Zebra to collect a rotation dataset. Any plan using this is
    responsible for setting the smargon velocity appropriately so that the desired
    image width is achieved with the exposure time given here.

    Parameters:
        zebra:              The zebra device to use
        axis:               Encoder enum representing which axis to use for position
                            compare. Currently always omega.
        start_angle:        Position at which the scan should begin, in degrees.
        scan_width:         Total angle through which to collect, in degrees.
        shutter_opening_deg:How many degrees of rotation it takes for the fast shutter
                            to open. Increases the gate width.
        shutter_opening_s:  How many seconds it takes for the fast shutter to open. The
                            detector pulse is delayed after the shutter signal by this
                            amount.
        direction:          RotationDirection enum for positive or negative.
                            Defaults to Positive.
        group:              A name for the group of statuses generated
        wait:               Block until all the settings have completed
        ttl_input_for_detector_to_use: If the zebra isn't using the TTL_DETECTOR zebra input,
                            manually specify which TTL input is being used for the desired detector
    """

    ttl_detector = ttl_input_for_detector_to_use or zebra.mapping.outputs.TTL_DETECTOR

    yield from bps.abs_set(zebra.pc.dir, direction.value, group=group)
    LOGGER.info("ZEBRA SETUP: START")
    # Set gate start, adjust for shutter opening time if necessary
    LOGGER.info(f"ZEBRA SETUP: degrees to adjust for shutter = {shutter_opening_deg}")
    LOGGER.info(f"ZEBRA SETUP: start angle start: {start_angle}")
    LOGGER.info(f"ZEBRA SETUP: start angle adjusted, gate start set to: {start_angle}")
    yield from bps.abs_set(zebra.pc.gate_start, start_angle, group=group)
    # set gate width to total width
    yield from bps.abs_set(
        zebra.pc.gate_width, scan_width + shutter_opening_deg, group=group
    )
    LOGGER.info(
        f"Pulse start set to shutter open time, set to: {abs(shutter_opening_s)}"
    )
    yield from bps.abs_set(zebra.pc.pulse_start, abs(shutter_opening_s), group=group)
    # Set gate position to be angle of interest
    yield from bps.abs_set(zebra.pc.gate_trigger, axis.value, group=group)
    # Set shutter to automatic and to trigger via PC_GATE
    yield from configure_zebra_and_shutter_for_auto_shutter(
        zebra, zebra_shutter, zebra.mapping.sources.PC_GATE, group=group
    )
    # Trigger the detector with a pulse
    yield from bps.abs_set(
        zebra.output.out_pvs[ttl_detector],
        zebra.mapping.sources.PC_PULSE,
        group=group,
    )

    LOGGER.info(f"ZEBRA SETUP: END - {'' if wait else 'not'} waiting for completion")
    if wait:
        yield from bps.wait(group, timeout=ZEBRA_STATUS_TIMEOUT)


def tidy_up_zebra_after_rotation_scan(
    zebra: Zebra,
    zebra_shutter: ZebraShutter,
    group="tidy_up_zebra_after_rotation",
    wait=True,
):
    """
    Set the zebra back to a state which is expected by GDA.

    Args:
        zebra: Zebra device.
        zebra_shutter: Zebra shutter device.
        group: Bluesky group to use when waiting on completion.
        wait: If true, block until completion.
    """

    yield from bps.abs_set(zebra.pc.arm, ArmDemand.DISARM, group=group)
    yield from bps.abs_set(
        zebra_shutter.control_mode, ZebraShutterControl.MANUAL, group=group
    )
    if wait:
        yield from bps.wait(group, timeout=ZEBRA_STATUS_TIMEOUT)



from pathlib import Path

import pytest

from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import (
    ChipType,
    MappingType,
    PumpProbeSetting,
)
from mx_bluesky.beamlines.i24.serial.parameters import (
    DetectorName,
    FixedTargetParameters,
    get_chip_format,
)

TEST_PATH = Path("tests/test_data/test_daq_configuration")

TEST_LUT = {
    DetectorName.EIGER: TEST_PATH / "lookup/test_det_dist_converter.txt",
}


@pytest.fixture
def dummy_params_with_pp():
    oxford_defaults = get_chip_format(ChipType.Oxford)
    return FixedTargetParameters(
        visit=Path("/tmp/dls/i24/fixed/foo"),
        directory="bar",
        filename="chip",
        exposure_time_s=0.01,
        detector_distance_mm=100,
        detector_name=DetectorName("eiger"),
        transmission=1.0,
        num_exposures=1,
        chip=oxford_defaults,
        map_type=MappingType(1),
        pump_repeat=PumpProbeSetting(3),
        checker_pattern=False,
        chip_map=[1],
        laser_dwell_s=0.02,
        laser_delay_s=0.05,
    )



from typing import Any

import pydantic
from bluesky import plan_stubs as bps
from bluesky.utils import MsgGenerator
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.backlight import Backlight
from dodal.devices.baton import Baton
from dodal.devices.detector.detector_motion import DetectorMotion, ShutterState
from dodal.devices.ipin import IPin, IPinGain
from dodal.devices.mx_phase1.beamstop import Beamstop, BeamstopPositions
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter, ZebraShutterState
from ophyd_async.core import InOut

from mx_bluesky.common.device_setup_plans.xbpm_feedback import (
    unpause_xbpm_feedback_and_set_transmission_to_1,
)
from mx_bluesky.common.utils.exceptions import BeamlineCheckFailureError
from mx_bluesky.common.utils.log import LOGGER

_GROUP_PRE_BEAMSTOP_OUT_CHECK = "pre_background_check"
_GROUP_POST_BEAMSTOP_OUT_CHECK = "post_background_check"
_PARAM_IPIN_THRESHOLD = "ipin_threshold"

_FEEDBACK_TIMEOUT_S = 10


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class BeamstopCheckDevices:
    aperture_scatterguard: ApertureScatterguard
    attenuator: BinaryFilterAttenuator
    backlight: Backlight
    baton: Baton
    beamstop: Beamstop
    detector_motion: DetectorMotion
    ipin: IPin
    sample_shutter: ZebraShutter
    xbpm_feedback: XBPMFeedback


class SampleCurrentBelowThresholdError(BeamlineCheckFailureError): ...


class BeamstopNotInPositionError(BeamlineCheckFailureError): ...


class BeamObstructedError(BeamlineCheckFailureError): ...


def move_beamstop_in_and_verify_using_diode(
    devices: BeamstopCheckDevices,
    beamline_parameters: dict[str, Any],
    detector_min_z_mm: float,
    detector_max_z_mm: float,
) -> MsgGenerator:
    """
    Move the beamstop into the data collection position, checking the beam current
    via the diode on the detector shutter, first with the beamstop out and then with
    the beamstop in.
    These checks aim to ensure that
       * The beam is not obstructed by something other than the beamstop
       * The beamstop has been successfully moved and is intercepting the beam

    As a side-effect, this plan also does the following things:
        * Move the detector z-axis in range
        * Move the backlight out
        * Unpauses feedback
        * Sets xmission to 100%
        * Sets IPin gain to 10^4 low noise
        * Moves aperture scatterguard to OUT_OF_BEAM if it is currently in beam

    Implementation note:
        Some checks are repeated here such as closing the sample shutter, so that at a
        future point this plan may be run independently of the udc default state script
        if desired.
    Note on commissioning mode:
        When commissioning mode is enabled, the beamstop check will execute normally except
        where it expects beam to be present, the absence of beam will be ignored.
    Args:
        devices: The device composite containing the necessary devices
        beamline_parameters: A mapping containing the beamlineParameters
        detector_min_z_mm: Detector minimum distance at which beamstop is effective
        detector_max_z_mm: Detector maximum distance at which beamstop is effective
    Raises:
        SampleCurrentBelowThresholdError: If we do not have sufficient sample current to perform
            the check.
        BeamstopNotInPositionError: If the ipin current is too high, indicating that the
            beamstop is not in the correct position.
        BeamObstructedError: If the ipin current is too low after the first check
            with the beamstop out, indicating a likely obstruction.
    """
    LOGGER.info("Performing beamstop check...")
    commissioning_mode_enabled = yield from bps.rd(devices.baton.commissioning)
    beamstop_threshold_uA = beamline_parameters[_PARAM_IPIN_THRESHOLD]  # noqa: N806

    yield from _start_moving_detector_if_needed(
        devices, detector_min_z_mm, detector_max_z_mm, _GROUP_POST_BEAMSTOP_OUT_CHECK
    )
    yield from _pre_beamstop_out_check_actions(devices)
    yield from _beamstop_out_check(
        devices, beamstop_threshold_uA, commissioning_mode_enabled
    )
    yield from _post_beamstop_out_check_actions(devices, _GROUP_POST_BEAMSTOP_OUT_CHECK)
    yield from _beamstop_in_check(devices, beamstop_threshold_uA)


def _pre_beamstop_out_check_actions(devices: BeamstopCheckDevices):
    # Re-verify that the sample shutter is closed
    yield from bps.abs_set(devices.sample_shutter, ZebraShutterState.CLOSE, wait=True)
    LOGGER.info("Unpausing feedback, transmission to 100%, wait for feedback stable...")
    try:
        yield from unpause_xbpm_feedback_and_set_transmission_to_1(
            devices.xbpm_feedback,
            devices.attenuator,
            _FEEDBACK_TIMEOUT_S,
        )
    except TimeoutError as e:
        raise SampleCurrentBelowThresholdError(
            "Unable to perform beamstop check - xbpm feedback did not become stable "
            " - check if beam present?"
        ) from e

    yield from bps.abs_set(
        devices.backlight, InOut.OUT, group=_GROUP_PRE_BEAMSTOP_OUT_CHECK
    )

    yield from bps.abs_set(
        devices.aperture_scatterguard.selected_aperture,
        ApertureValue.OUT_OF_BEAM,
        group=_GROUP_PRE_BEAMSTOP_OUT_CHECK,
    )

    yield from bps.abs_set(
        devices.ipin.gain,
        IPinGain.GAIN_10E4_LOW_NOISE,
        group=_GROUP_PRE_BEAMSTOP_OUT_CHECK,
    )
    yield from bps.abs_set(
        devices.detector_motion.shutter,
        ShutterState.CLOSED,
        group=_GROUP_PRE_BEAMSTOP_OUT_CHECK,
    )
    yield from bps.abs_set(
        devices.beamstop.selected_pos,
        BeamstopPositions.OUT_OF_BEAM,
        group=_GROUP_PRE_BEAMSTOP_OUT_CHECK,
    )

    LOGGER.info("Waiting for pre-background-check motions to complete...")
    yield from bps.wait(group=_GROUP_PRE_BEAMSTOP_OUT_CHECK)

    # Check detector shutter is closed
    detector_shutter_state = yield from bps.rd(devices.detector_motion.shutter)
    detector_shutter_is_closed = detector_shutter_state == ShutterState.CLOSED
    if not detector_shutter_is_closed:
        raise RuntimeError(
            "Unable to proceed with beamstop background check, detector shutter did not close"
        )


def _start_moving_detector_if_needed(
    devices: BeamstopCheckDevices,
    detector_min_z_mm: float,
    detector_max_z_mm: float,
    group: str | None = None,
):
    detector_current_z = yield from bps.rd(devices.detector_motion.z)
    target_z = max(min(detector_current_z, detector_max_z_mm), detector_min_z_mm)
    if detector_current_z != target_z:
        LOGGER.info(
            f"Detector distance {detector_current_z}mm outside acceptable range for diode "
            f"check {detector_min_z_mm} <= z <= {detector_max_z_mm}, moving it."
        )
        yield from bps.abs_set(devices.detector_motion.z, target_z, group=group)


def _post_beamstop_out_check_actions(devices: BeamstopCheckDevices, group: str):
    yield from bps.abs_set(
        devices.beamstop.selected_pos,
        BeamstopPositions.DATA_COLLECTION,
        group=group,
    )

    LOGGER.info("Waiting for detector motion to complete...")
    yield from bps.wait(group=group)


def _beamstop_out_check(
    devices: BeamstopCheckDevices,
    beamstop_threshold_uA: float,  # noqa: N803
    commissioning_mode_enabled: bool,
):
    ipin_beamstop_out_uA = yield from _check_ipin(devices)  # noqa: N806

    LOGGER.info(f"Beamstop out ipin = {ipin_beamstop_out_uA}uA")
    if ipin_beamstop_out_uA < beamstop_threshold_uA:
        msg = (
            f"IPin current {ipin_beamstop_out_uA}uA below threshold "
            f"{beamstop_threshold_uA} with beamstop out - check "
            f"that beam is not obstructed."
        )
        if commissioning_mode_enabled:
            LOGGER.warning(msg + " - commissioning mode enabled - ignoring this")
        else:
            raise BeamObstructedError(msg)


def _beamstop_in_check(devices: BeamstopCheckDevices, beamstop_threshold_uA: float):  # noqa: N803
    ipin_in_beam_uA = yield from _check_ipin(devices)  # noqa: N806
    if ipin_in_beam_uA > beamstop_threshold_uA:
        raise BeamstopNotInPositionError(
            f"Ipin is too high at {ipin_in_beam_uA} - check that beamstop is "
            f"in the correct position."
        )


def _check_ipin(devices: BeamstopCheckDevices):
    try:
        yield from bps.abs_set(
            devices.sample_shutter, ZebraShutterState.OPEN, wait=True
        )
        yield from bps.sleep(1)  # wait for reading to settle

        return (yield from bps.rd(devices.ipin.pin_readback))  # noqa: N806
    finally:
        yield from bps.abs_set(
            devices.sample_shutter, ZebraShutterState.CLOSE, wait=True
        )



from pathlib import Path

# import pytest
from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import (
    ChipType,
    MappingType,
    PumpProbeSetting,
)
from mx_bluesky.beamlines.i24.serial.parameters.constants import DetectorName
from mx_bluesky.beamlines.i24.serial.parameters.experiment_parameters import (
    ChipDescription,
    ExtruderParameters,
    FixedTargetParameters,
    SSXType,
)
from mx_bluesky.beamlines.i24.serial.parameters.utils import get_chip_format


def test_extruder_params(dummy_params_ex):
    assert isinstance(dummy_params_ex, ExtruderParameters)
    assert dummy_params_ex.collection_directory == Path("/tmp/dls/i24/extruder/foo/bar")
    assert dummy_params_ex.nexgen_experiment_type == "extruder"
    assert dummy_params_ex.ispyb_experiment_type is SSXType.EXTRUDER


def test_chip_params_with_mapping_lite_and_one_block(dummy_params_without_pp):
    assert isinstance(dummy_params_without_pp, FixedTargetParameters)
    assert dummy_params_without_pp.total_num_images == 400
    assert dummy_params_without_pp.chip.chip_type is ChipType.Oxford
    assert dummy_params_without_pp.nexgen_experiment_type == "fixed-target"
    assert dummy_params_without_pp.ispyb_experiment_type is SSXType.FIXED


def test_chip_params_with_multiple_blocks(dummy_params_without_pp):
    dummy_params_without_pp.chip_map = [1, 2, 3]
    assert dummy_params_without_pp.total_num_images == 1200
    assert dummy_params_without_pp.map_type is MappingType.Lite
    assert dummy_params_without_pp.pump_repeat is PumpProbeSetting.NoPP


def test_chip_params_with_multiple_exposures(dummy_params_without_pp):
    dummy_params_without_pp.num_exposures = 2
    dummy_params_without_pp.chip_map = [1, 16]
    assert dummy_params_without_pp.total_num_images == 1600


def test_chip_params_with_no_mapping_for_oxford_chip():
    oxford_defaults = get_chip_format(ChipType.Oxford)
    oxford_chip_params = FixedTargetParameters(
        visit=Path("foo"),
        directory="bar",
        filename="chip",
        exposure_time_s=0.01,
        detector_distance_mm=100,
        detector_name=DetectorName("eiger"),
        transmission=1.0,
        num_exposures=1,
        chip=oxford_defaults,
        map_type=MappingType(0),
        pump_repeat=PumpProbeSetting(0),
        checker_pattern=False,
        chip_map=[],
    )
    assert oxford_chip_params.total_num_images == 25600


def test_chip_params_with_no_mapping_for_custom_chip():
    custom_defaults = ChipDescription(
        chip_type=ChipType.Custom,
        x_num_steps=5,
        y_num_steps=3,
        x_step_size=0.1,
        y_step_size=0.1,
        x_blocks=1,
        y_blocks=1,
        b2b_horz=0.0,
        b2b_vert=0.0,
    )
    custom_chip_params = FixedTargetParameters(
        visit=Path("foo"),
        directory="bar",
        filename="chip",
        exposure_time_s=0.01,
        detector_distance_mm=100,
        detector_name=DetectorName("eiger"),
        transmission=1.0,
        num_exposures=2,
        chip=custom_defaults,
        map_type=MappingType(0),
        pump_repeat=PumpProbeSetting(0),
        checker_pattern=False,
        chip_map=[],
    )
    assert custom_chip_params.total_num_images == 30



from unittest.mock import MagicMock, patch

import pytest
from bluesky import RunEngine
from dodal.beamlines import i24
from dodal.devices.attenuator.attenuator import EnumFilterAttenuator
from dodal.devices.beamlines.i24.aperture import Aperture
from dodal.devices.beamlines.i24.beamstop import Beamstop
from dodal.devices.beamlines.i24.commissioning_jungfrau import CommissioningJungfrau
from dodal.devices.beamlines.i24.dcm import DCM
from dodal.devices.beamlines.i24.dual_backlight import DualBacklight
from dodal.devices.beamlines.i24.vgonio import VerticalGoniometer
from dodal.devices.hutch_shutter import InterlockedHutchShutter
from dodal.devices.motors import YZStage
from dodal.devices.synchrotron import Synchrotron
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra import Zebra
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter

from mx_bluesky.beamlines.i24.jungfrau_commissioning.composites import (
    RotationScanComposite,
)
from mx_bluesky.beamlines.i24.web_gui_plans.jungfrau_plans import (
    gui_run_jf_rotation_scan,
)


@pytest.fixture
def vertical_gonio() -> VerticalGoniometer:
    return i24.vgonio.build(connect_immediately=True, mock=True)


def test_run_jf_rotation(
    jungfrau: CommissioningJungfrau,
    zebra: Zebra,
    enum_attenuator: EnumFilterAttenuator,
    aperture: Aperture,
    vertical_gonio: VerticalGoniometer,
    beamstop: Beamstop,
    detector_stage: YZStage,
    backlight: DualBacklight,
    dcm: DCM,
    synchrotron: Synchrotron,
    shutter: InterlockedHutchShutter,
    sample_shutter: ZebraShutter,
    xbpm_feedback: XBPMFeedback,
    run_engine: RunEngine,
):
    composite = RotationScanComposite(
        aperture,
        enum_attenuator,
        jungfrau,
        vertical_gonio,
        synchrotron,
        sample_shutter,
        zebra,
        xbpm_feedback,
        shutter,
        beamstop,
        detector_stage,
        backlight,
        dcm,
    )
    with patch(
        "mx_bluesky.beamlines.i24.web_gui_plans.jungfrau_plans.rotation_scan_plan",
        MagicMock(return_value=iter([])),
    ) as patch_inner_plan:
        run_engine(
            gui_run_jf_rotation_scan(
                "new_rotation", 0.01, 0.0, 0.1, 300, 1, [0.3], composite
            )
        )

        patch_inner_plan.assert_called_once()



from unittest.mock import MagicMock, patch

import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.beamlines.i23 import I23DetectorPositions
from dodal.devices.motors import SixAxisGonio
from dodal.devices.positioner import Positioner1D
from ophyd_async.core import get_mock_put, init_devices

from mx_bluesky.beamlines.i23.serial import one_nd_step, serial_collection


@pytest.fixture
def mock_gonio():
    with init_devices(mock=True):
        gonio = SixAxisGonio("", name="gonio")
    return gonio


@pytest.fixture
def mock_detector_motion():
    with init_devices(mock=True):
        detector_motion = Positioner1D("", I23DetectorPositions)
    return detector_motion


def test_when_grid_scan_called_then_expected_x_y_set(
    sim_run_engine: RunEngineSimulator,
    mock_detector_motion: Positioner1D,
    mock_gonio: SixAxisGonio,
):
    msgs = sim_run_engine.simulate_plan(
        serial_collection(4, 4, 0.1, 0.1, 30, 1.0, mock_detector_motion, mock_gonio)
    )
    x_moves = [
        msg for msg in msgs if msg.command == "set" and msg.obj.name == "gonio-x"
    ]
    y_moves = [
        msg for msg in msgs if msg.command == "set" and msg.obj.name == "gonio-y"
    ]
    assert len(x_moves) == 4 * 4 + 1  # Additional 1 for the initial move
    assert len(y_moves) == 4 * 4 + 1

    assert x_moves[1].args[0] == pytest.approx(0.1)
    assert y_moves[5].args[0] == pytest.approx(0.1)


def test_omega_moves_twice_for_every_point(
    sim_run_engine: RunEngineSimulator,
    mock_detector_motion: Positioner1D,
    mock_gonio: SixAxisGonio,
):
    msgs = sim_run_engine.simulate_plan(
        serial_collection(4, 4, 0.1, 0.1, 30, 1.0, mock_detector_motion, mock_gonio)
    )
    omega_moves = [
        msg for msg in msgs if msg.command == "set" and msg.obj.name == "gonio-omega"
    ]
    assert len(omega_moves) == (4 * 4) * 2


@patch("mx_bluesky.beamlines.i23.serial.short_uid")
def test_when_stopped_at_point_then_omega_rotated_at_expected_speed(
    mock_short_uid: MagicMock,
    sim_run_engine: RunEngineSimulator,
    mock_gonio: SixAxisGonio,
):
    mock_short_uid.return_value = "test"
    msgs = sim_run_engine.simulate_plan(
        one_nd_step(
            [],
            {mock_gonio.x: 0.1, mock_gonio.y: 0.1},
            MagicMock(),
            mock_gonio.omega,
            30.0,
            1.0,
        )
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "wait" and msg.kwargs["group"] == "test"
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "gonio-omega-velocity"
        and msg.args[0] == 1.0,
    )
    msgs = assert_message_and_return_remaining(msgs, lambda msg: msg.command == "wait")
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "gonio-omega"
        and msg.args[0] == 30.0,
    )


@patch("mx_bluesky.beamlines.i23.serial.short_uid")
def test_omega_set_to_0_at_max_velo_during_grid_move(
    mock_short_uid: MagicMock,
    sim_run_engine: RunEngineSimulator,
    mock_gonio: SixAxisGonio,
):
    mock_short_uid.return_value = "test"
    sim_run_engine.add_read_handler_for(mock_gonio.omega.max_velocity, 90)
    msgs = sim_run_engine.simulate_plan(
        one_nd_step(
            [],
            {mock_gonio.x: 0.1, mock_gonio.y: 0.1},
            MagicMock(),
            mock_gonio.omega,
            30.0,
            1.0,
        )
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "set" and msg.obj.name == "gonio-x"
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "gonio-omega-velocity"
        and msg.args[0] == 90,
    )
    msgs = assert_message_and_return_remaining(msgs, lambda msg: msg.command == "wait")
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "gonio-omega"
        and msg.args[0] == 0,
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "wait" and msg.kwargs["group"] == "test"
    )


def test_detector_moves_in_at_experiment_start(
    sim_run_engine: RunEngineSimulator,
    mock_detector_motion: Positioner1D,
    mock_gonio: SixAxisGonio,
):
    msgs = sim_run_engine.simulate_plan(
        serial_collection(4, 4, 0.1, 0.1, 30, 1.0, mock_detector_motion, mock_gonio)
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "detector_motion-stage_position"
        and msg.args[0] == "In",
    )


async def test_serial_collection_can_run_in_real_run_engine(
    run_engine: RunEngine, mock_detector_motion: Positioner1D, mock_gonio: SixAxisGonio
):
    run_engine(
        serial_collection(4, 4, 0.1, 0.1, 30, 1.0, mock_detector_motion, mock_gonio)
    )
    assert get_mock_put(mock_gonio.x.user_setpoint).call_count == 4 * 4 + 1
    assert get_mock_put(mock_gonio.y.user_setpoint).call_count == 4 * 4 + 1



from bluesky.utils import MsgGenerator
from dodal.common import inject
from dodal.devices.beamlines.i24.commissioning_jungfrau import CommissioningJungfrau
from ophyd_async.core import (
    WatchableAsyncStatus,
)
from ophyd_async.fastcs.jungfrau import (
    GainMode,
    create_jungfrau_external_triggering_info,
)
from pydantic import PositiveInt

from mx_bluesky.beamlines.i24.jungfrau_commissioning.plan_stubs.plan_utils import (
    fly_jungfrau,
)


def do_external_acquisition(
    exp_time_s: float,
    gain_mode: GainMode,
    total_triggers: PositiveInt = 1,
    wait: bool = False,
    jungfrau: CommissioningJungfrau = inject("commissioning_jungfrau"),
) -> MsgGenerator[WatchableAsyncStatus]:
    """
    Kickoff external triggering on the Jungfrau, and optionally wait for completion.

    Any plan using this stub MUST stage the Jungfrau with the stage_decorator and open a run,
    ideally using the run_decorator.

    Args:
        exp_time_s: Length of detector exposure for each frame.
        total_triggers: Number of external triggers received before acquisition is marked as complete.
        jungfrau: Jungfrau device
        wait: Optionally block until data collection is complete.
    """

    trigger_info = create_jungfrau_external_triggering_info(total_triggers, exp_time_s)
    status = yield from fly_jungfrau(jungfrau, trigger_info, gain_mode, wait=wait)
    return status



import re

import bluesky.preprocessors as bpp
import numpy as np
import pytest
from bluesky.run_engine import RunEngine
from dodal.devices.eiger import EigerDetector
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.smargon import Smargon
from dodal.devices.zocalo import ZocaloResults
from dodal.utils import is_test_mode

from mx_bluesky.common.experiment_plans.inner_plans.read_hardware import (
    read_hardware_for_zocalo,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanPlane,
    ispyb_activation_wrapper,
)
from mx_bluesky.common.parameters.constants import (
    EnvironmentConstants,
    PlanNameConstants,
)
from mx_bluesky.hyperion.external_interaction.callbacks.__main__ import (
    create_gridscan_callbacks,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan
from tests.conftest import create_dummy_scan_spec

"""
If fake-zocalo system tests are failing, check that the RMQ instance is set up right:

- Open the RMQ webpage specified when you start the fake zocalo and login with the
provided credentials

- go to the admin panel and under the 'exchanges' tab ensure that there is a 'results'
exchange for the zocalo vhost (other settings can be left blank)

- go to the 'queues and streams' tab and add a binding for the xrc.i03 queue to the
results exchange, with the routing key 'xrc.i03'

- make sure that there are no un-acked/un-delivered messages on the i03.xrc queue
"""


@bpp.set_run_key_decorator("testing125")
@bpp.run_decorator(
    md={
        "subplan_name": PlanNameConstants.DO_FGS,
        "zocalo_environment": EnvironmentConstants.ZOCALO_ENV,
        "omega_to_scan_spec": {
            GridscanPlane.OMEGA_XY: create_dummy_scan_spec()[0],
            GridscanPlane.OMEGA_XZ: create_dummy_scan_spec()[1],
        },
    }
)
def fake_fgs_plan(eiger: EigerDetector):
    yield from read_hardware_for_zocalo(eiger)


@pytest.fixture
def run_zocalo_with_dev_ispyb(
    dummy_params: HyperionSpecifiedThreeDGridScan,
    dummy_ispyb,
    run_engine: RunEngine,
    zocalo_for_fake_zocalo: ZocaloResults,
    eiger: EigerDetector,
    oav_for_system_test: OAV,
    smargon: Smargon,
    fake_grid_snapshot_plan,
):
    async def inner(sample_name="", fallback=np.array([0, 0, 0])):
        dummy_params.file_name = sample_name
        _, ispyb_callback = create_gridscan_callbacks()
        run_engine.subscribe(ispyb_callback)

        def trigger_zocalo_after_fast_grid_scan():
            yield from fake_grid_snapshot_plan(smargon, oav_for_system_test)

            @bpp.set_run_key_decorator("testing124")
            @bpp.stage_decorator([zocalo_for_fake_zocalo])
            @bpp.run_decorator(
                md={
                    "subplan_name": CONST.PLAN.GRIDSCAN_OUTER,
                    "zocalo_environment": EnvironmentConstants.ZOCALO_ENV,
                    "mx_bluesky_parameters": dummy_params.model_dump_json(),
                }
            )
            def inner_plan():
                yield from fake_fgs_plan(eiger)

            yield from inner_plan()

        run_engine(
            ispyb_activation_wrapper(
                trigger_zocalo_after_fast_grid_scan(), dummy_params
            )
        )
        centre = await zocalo_for_fake_zocalo.centre_of_mass.get_value()
        if centre.size == 0:
            centre = fallback
        else:
            centre = centre[0]

        return ispyb_callback, ispyb_callback.emit_cb, centre

    return inner


@pytest.mark.system_test
def test_is_test_mode():
    assert is_test_mode(), "DODAL_TEST_MODE must be set in environment before launch"


@pytest.mark.system_test
async def test_given_a_result_with_no_diffraction_when_zocalo_called_then_move_to_fallback(
    run_zocalo_with_dev_ispyb,
):
    fallback = np.array([1, 2, 3])
    _, _, centre = await run_zocalo_with_dev_ispyb("NO_DIFF", fallback)
    assert np.allclose(centre, fallback)


@pytest.mark.system_test
async def test_zocalo_adds_nonzero_comment_time(
    run_zocalo_with_dev_ispyb, fetch_comment
):
    ispyb, zc, _ = await run_zocalo_with_dev_ispyb()

    comment = fetch_comment(ispyb.ispyb_ids.data_collection_ids[0])
    match = re.match(r".*Zocalo processing took (\d+\.\d+) s", comment)
    assert match
    time_s = float(match.group(1))
    assert time_s > 0
    assert time_s < 180



"""
This module contains the bluesky plan entry points for use with hyperion.
hyperion-blueapi loads this module via blueapi.py which contains module-load initialisation
hooks that are run when blueapi loads the plan module, which are not needed by the
in-process runner and so not defined here.

The json schema and documentation therein generated by the blueapi /plans endpoint
from this file constitutes the hyperion-blueapi interface to the hyperion supervisor
process.
"""

from bluesky import plan_stubs as bps
from bluesky.utils import MsgGenerator
from dodal.common import inject
from dodal.devices.aperturescatterguard import ApertureScatterguard
from dodal.devices.detector.detector_motion import DetectorMotion, ShutterState
from dodal.devices.motors import XYZStage
from dodal.devices.robot import BartRobot
from dodal.devices.smargon import Smargon

from mx_bluesky.common.device_setup_plans.robot_load_unload import (
    robot_unload as _robot_unload,
)
from mx_bluesky.hyperion.blueapi.parameters import (
    LoadCentreCollectParams,
    load_centre_collect_to_internal,
)
from mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan import (
    LoadCentreCollectComposite,
)
from mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan import (
    load_centre_collect_full as _load_centre_collect_full,
)
from mx_bluesky.hyperion.experiment_plans.udc_default_state import (
    UDCDefaultDevices,
)
from mx_bluesky.hyperion.experiment_plans.udc_default_state import (
    move_to_udc_default_state as _move_to_udc_default_state,
)

__all__ = [
    "LoadCentreCollectComposite",
    "LoadCentreCollectParams",
    "UDCDefaultDevices",
    "clean_up_udc",
    "load_centre_collect",
    "move_to_udc_default_state",
    "robot_unload",
]


def load_centre_collect(
    parameters: LoadCentreCollectParams,
    composite: LoadCentreCollectComposite = inject(),
) -> MsgGenerator:
    """
    Attempt a complete data collection experiment, consisting of the following:
        * Load the sample if necessary
        * Move to the specified goniometer start angles
        * Perform optical centring, then X-ray centring
        * If X-ray centring finds one or more diffracting centres then for each centre
    that satisfies the chosen selection function,
    move to that centre and do a collection with the specified parameters.
    """
    yield from _load_centre_collect_full(
        composite, load_centre_collect_to_internal(parameters)
    )


def robot_unload(
    visit: str,
    robot: BartRobot = inject("robot"),
    smargon: Smargon = inject("gonio"),
    aperture_scatterguard: ApertureScatterguard = inject("aperture_scatterguard"),
    lower_gonio: XYZStage = inject("lower_gonio"),
) -> MsgGenerator:
    """
    Unload the currently mounted pin into the location that it was loaded from.
    This is to be invoked as the final step upon successful completion of the UDC queue.
    """
    yield from _robot_unload(robot, smargon, aperture_scatterguard, lower_gonio, visit)


def clean_up_udc(
    cleanup_group: str = "cleanup",
    detector_motion: DetectorMotion = inject("detector_motion"),
) -> MsgGenerator:
    """
    Perform actions that are always performed on ending UDC, regardless of whether
    it terminated abnormally or not.
    Currently the only thing this does is close the detector shutter.
    Args:
        detector_motion:

    Returns:
    """
    yield from bps.abs_set(
        detector_motion.shutter, ShutterState.CLOSED, group=cleanup_group
    )
    yield from bps.wait(cleanup_group)


def move_to_udc_default_state(
    composite: UDCDefaultDevices = inject(),
) -> MsgGenerator:
    """
    Move beamline hardware to known positions prior to UDC start.
    """
    yield from _move_to_udc_default_state(composite)



from mx_bluesky.common.parameters.rotation import SingleRotationScan
from tests.conftest import raw_params_from_file


def get_good_single_rotation_params(tmp_path):
    params = raw_params_from_file(
        "tests/unit_tests/beamlines/i24/jungfrau_commissioning/test_data/test_good_rotation_params.json",
        tmp_path,
    )

    return SingleRotationScan(**params)



import logging
from abc import abstractmethod
from collections.abc import Callable
from contextlib import AbstractContextManager
from threading import Thread
from time import sleep  # noqa
from urllib import request
from urllib.error import URLError

from blueapi.config import ApplicationConfig, ConfigLoader
from bluesky.callbacks import CallbackBase
from bluesky.callbacks.zmq import Proxy, RemoteDispatcher
from bluesky_stomp.messaging import StompClient
from bluesky_stomp.models import Broker
from dodal.log import LOGGER as DODAL_LOGGER
from dodal.log import set_up_all_logging_handlers

from mx_bluesky.common.external_interaction.alerting import set_alerting_service
from mx_bluesky.common.external_interaction.alerting.log_based_service import (
    LoggingAlertService,
)
from mx_bluesky.common.external_interaction.callbacks.common.log_uid_tag_callback import (
    LogUidTaggingCallback,
)
from mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback import (
    ZocaloCallback,
)
from mx_bluesky.common.external_interaction.callbacks.sample_handling.sample_handling_callback import (
    SampleHandlingCallback,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanISPyBCallback,
    generate_start_info_from_omega_map,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback import (
    GridscanNexusFileCallback,
)
from mx_bluesky.common.utils.log import (
    ISPYB_ZOCALO_CALLBACK_LOGGER,
    NEXUS_LOGGER,
    _get_logging_dirs,
    tag_filter,
)
from mx_bluesky.hyperion.external_interaction.callbacks.alert_on_container_change import (
    AlertOnContainerChange,
)
from mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback import (
    RobotLoadISPyBCallback,
)
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback import (
    RotationISPyBCallback,
    generate_start_info_from_ordered_runs,
)
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.nexus_callback import (
    RotationNexusFileCallback,
)
from mx_bluesky.hyperion.external_interaction.callbacks.snapshot_callback import (
    BeamDrawingCallback,
)
from mx_bluesky.hyperion.external_interaction.callbacks.stomp.dispatcher import (
    StompDispatcher,
)
from mx_bluesky.hyperion.parameters.cli import CallbackArgs, parse_callback_args
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.gridscan import (
    GridCommonWithHyperionDetectorParams,
    HyperionSpecifiedThreeDGridScan,
)

PING_TIMEOUT_S = 1

LIVENESS_POLL_SECONDS = 1
ERROR_LOG_BUFFER_LINES = 5000
HYPERION_PING_INTERVAL_S = 19


def create_gridscan_callbacks() -> tuple[
    GridscanNexusFileCallback, GridscanISPyBCallback
]:
    return (
        GridscanNexusFileCallback(param_type=HyperionSpecifiedThreeDGridScan),
        GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams,
            emit=ZocaloCallback(
                CONST.PLAN.DO_FGS, CONST.ZOCALO_ENV, generate_start_info_from_omega_map
            ),
        ),
    )


def create_rotation_callbacks() -> tuple[
    RotationNexusFileCallback, RotationISPyBCallback
]:
    return (
        RotationNexusFileCallback(),
        RotationISPyBCallback(
            emit=ZocaloCallback(
                CONST.PLAN.ROTATION_MULTI,
                CONST.ZOCALO_ENV,
                generate_start_info_from_ordered_runs,
            )
        ),
    )


def setup_callbacks() -> list[CallbackBase]:
    rot_nexus_cb, rot_ispyb_cb = create_rotation_callbacks()
    snapshot_cb = BeamDrawingCallback(emit=rot_ispyb_cb)
    return [
        *create_gridscan_callbacks(),
        rot_nexus_cb,
        snapshot_cb,
        LogUidTaggingCallback(),
        RobotLoadISPyBCallback(),
        SampleHandlingCallback(),
        AlertOnContainerChange(),
    ]


def setup_logging(dev_mode: bool):
    for logger, filename in [
        (ISPYB_ZOCALO_CALLBACK_LOGGER, "hyperion_ispyb_callback.log"),
        (NEXUS_LOGGER, "hyperion_nexus_callback.log"),
    ]:
        logging_path, debug_logging_path = _get_logging_dirs(dev_mode)
        if logger.handlers == []:
            handlers = set_up_all_logging_handlers(
                logger,
                logging_path,
                filename,
                dev_mode,
                ERROR_LOG_BUFFER_LINES,
                CONST.GRAYLOG_PORT,
                debug_logging_path,
            )
            handlers["graylog_handler"].addFilter(tag_filter)
    log_info(f"Loggers initialised with dev_mode={dev_mode}")
    nexgen_logger = logging.getLogger("nexgen")
    nexgen_logger.parent = NEXUS_LOGGER
    DODAL_LOGGER.parent = ISPYB_ZOCALO_CALLBACK_LOGGER
    log_debug("nexgen logger added to nexus logger")


def log_info(msg, *args, **kwargs):
    ISPYB_ZOCALO_CALLBACK_LOGGER.info(msg, *args, **kwargs)
    NEXUS_LOGGER.info(msg, *args, **kwargs)


def log_debug(msg, *args, **kwargs):
    ISPYB_ZOCALO_CALLBACK_LOGGER.debug(msg, *args, **kwargs)
    NEXUS_LOGGER.debug(msg, *args, **kwargs)


class HyperionCallbackRunner:
    """Runs Nexus, ISPyB and Zocalo callbacks in their own process."""

    def __init__(self, callback_args: CallbackArgs) -> None:
        setup_logging(callback_args.dev_mode)
        log_info("Hyperion callback process started.")
        set_alerting_service(LoggingAlertService(CONST.GRAYLOG_STREAM_ID))

        self.callbacks = setup_callbacks()

        self.watchdog_thread = Thread(
            target=run_watchdog,
            daemon=True,
            name="Watchdog",
            args=[callback_args.watchdog_port],
        )

        self._dispatcher_cm: DispatcherContextMgr
        if callback_args.stomp_config:
            self._dispatcher_cm = StompDispatcherContextMgr(
                callback_args, self.callbacks
            )
        else:
            self._dispatcher_cm = RemoteDispatcherContextMgr(self.callbacks)

    def start(self):
        log_info(f"Launching threads, with callbacks: {self.callbacks}")
        self.watchdog_thread.start()
        with self._dispatcher_cm:
            ping_watchdog_while_alive(self._dispatcher_cm, self.watchdog_thread)


def run_watchdog(watchdog_port: int):
    log_info("Hyperion watchdog keepalive running")
    while True:
        try:
            with request.urlopen(
                f"http://localhost:{watchdog_port}/callbackPing",
                timeout=PING_TIMEOUT_S,
            ) as response:
                if response.status != 200:
                    log_debug(
                        f"Unable to ping Hyperion liveness endpoint, status {response.status}"
                    )
        except URLError as e:
            log_debug("Unable to ping Hyperion liveness endpoint", exc_info=e)
        sleep(HYPERION_PING_INTERVAL_S)


def main(dev_mode=False) -> None:
    callback_args = parse_callback_args()
    callback_args.dev_mode = dev_mode or callback_args.dev_mode
    print(f"In dev mode: {dev_mode}")
    runner = HyperionCallbackRunner(callback_args)
    runner.start()


class DispatcherContextMgr(AbstractContextManager):
    @abstractmethod
    def is_alive(self) -> bool: ...


class RemoteDispatcherContextMgr(DispatcherContextMgr):
    def __init__(self, callbacks: list[CallbackBase]):
        super().__init__()

        self.proxy = Proxy(*CONST.CALLBACK_0MQ_PROXY_PORTS)
        self.proxy_thread = Thread(
            target=self.proxy.start, daemon=True, name="0MQ Proxy"
        )

        self.dispatcher = RemoteDispatcher(
            f"localhost:{CONST.CALLBACK_0MQ_PROXY_PORTS[1]}"
        )

        def start_dispatcher(callbacks: list[Callable]):
            for cb in callbacks:
                self.dispatcher.subscribe(cb)
            self.dispatcher.start()

        self.dispatcher_thread = Thread(
            target=start_dispatcher,
            args=[callbacks],
            daemon=True,
            name="0MQ Dispatcher",
        )
        log_info("Created 0MQ proxy and local RemoteDispatcher.")

    def __enter__(self):
        log_info("Proxy and dispatcher thread launched.")
        self.proxy_thread.start()
        self.dispatcher_thread.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback, /):
        self.dispatcher.stop()
        # proxy has no way to stop

    def is_alive(self):
        return self.proxy_thread.is_alive() and self.dispatcher_thread.is_alive()


class StompDispatcherContextMgr(DispatcherContextMgr):
    def __init__(self, args: CallbackArgs, callbacks: list[CallbackBase]):
        super().__init__()
        loader = ConfigLoader(ApplicationConfig)
        assert args.stomp_config is not None
        loader.use_values_from_yaml(args.stomp_config)
        config = loader.load()
        log_info(
            f"Stomp client configured on {config.stomp.url.host}:{config.stomp.url.port}"
        )
        assert config.stomp.url.host is not None
        assert config.stomp.url.port is not None
        self._stomp_client = StompClient.for_broker(
            broker=Broker(
                host=config.stomp.url.host,
                port=config.stomp.url.port,
                auth=config.stomp.auth,
            )
        )
        self._dispatcher = StompDispatcher(self._stomp_client)
        for cb in callbacks:
            self._dispatcher.subscribe(cb)

    def is_alive(self) -> bool:
        return self._stomp_client.is_connected()

    def __enter__(self):
        self._dispatcher.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback, /):
        self._dispatcher.__exit__(exc_type, exc_value, traceback)


def ping_watchdog_while_alive(
    dispatcher_cm: DispatcherContextMgr, watchdog_thread: Thread
):
    alive = watchdog_thread.is_alive() and dispatcher_cm.is_alive()
    try:
        log_debug("Trying to wait forever on callback and dispatcher threads")
        while alive:
            sleep(LIVENESS_POLL_SECONDS)
            alive = watchdog_thread.is_alive() and dispatcher_cm.is_alive()
    except KeyboardInterrupt:
        log_info("Main thread received interrupt - exiting.")
    else:
        log_info("Proxy or dispatcher thread ended - exiting.")


if __name__ == "__main__":
    main()



from unittest.mock import patch

import pytest
from dodal.devices.beamlines.i24.aperture import Aperture
from dodal.devices.beamlines.i24.beam_center import DetectorBeamCenter
from dodal.devices.beamlines.i24.beamstop import Beamstop
from dodal.devices.beamlines.i24.dual_backlight import DualBacklight
from dodal.devices.motors import YZStage
from ophyd_async.core import set_mock_value

from mx_bluesky.beamlines.i24.serial.setup_beamline import setup_beamline

from ..conftest import TEST_LUT


@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.bps.sleep")
async def test_setup_beamline_for_collection_plan(
    _, aperture: Aperture, backlight: DualBacklight, beamstop: Beamstop, run_engine
):
    run_engine(
        setup_beamline.setup_beamline_for_collection_plan(aperture, backlight, beamstop)
    )

    assert await aperture.position.get_value() == "In"
    assert await beamstop.pos_select.get_value() == "Data Collection"
    assert await beamstop.y_rotation.user_setpoint.get_value() == 0

    assert await backlight.backlight_position.pos_level.get_value() == "Out"


async def test_move_detector_stage_to_position_plan(
    detector_stage: YZStage, run_engine
):
    det_dist = 100
    run_engine(
        setup_beamline.move_detector_stage_to_position_plan(detector_stage, det_dist)
    )

    assert await detector_stage.z.user_setpoint.get_value() == det_dist


def test_compute_beam_center_position_from_lut(dummy_params_ex):
    lut_path = TEST_LUT[dummy_params_ex.detector_name]

    expected_beam_x = 1597.06
    expected_beam_y = 1693.33

    beam_center_pos = setup_beamline.compute_beam_center_position_from_lut(
        lut_path,
        dummy_params_ex.detector_distance_mm,
        dummy_params_ex.detector_size_constants,
    )
    assert beam_center_pos[0] == pytest.approx(expected_beam_x, 1e-2)
    assert beam_center_pos[1] == pytest.approx(expected_beam_y, 1e-2)


async def test_set_detector_beam_center_plan(
    eiger_beam_center: DetectorBeamCenter, dummy_params_ex, run_engine
):
    beam_center_pos = setup_beamline.compute_beam_center_position_from_lut(
        TEST_LUT[dummy_params_ex.detector_name],
        dummy_params_ex.detector_distance_mm,  # 100
        dummy_params_ex.detector_size_constants,
    )
    # test_detector_distance = 100
    # test_detector_params = dummy_params_ex.detector_params
    run_engine(
        setup_beamline.set_detector_beam_center_plan(
            eiger_beam_center,
            beam_center_pos,  # test_detector_params, test_detector_distance
        )
    )

    assert await eiger_beam_center.beam_x.get_value() == pytest.approx(1597.06, 1e-2)
    assert await eiger_beam_center.beam_y.get_value() == pytest.approx(1693.33, 1e-2)


@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caput")
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caget")
def test_eiger_raises_error_if_quickshot_and_no_args_list(
    fake_caget, fake_caput, run_engine, dcm, detector_stage
):
    with pytest.raises(TypeError):
        run_engine(setup_beamline.eiger("quickshot", None, dcm, detector_stage))


@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caput")
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caget")
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.bps.sleep")
def test_eiger_quickshot(_, fake_caget, fake_caput, run_engine, dcm, detector_stage):
    run_engine(
        setup_beamline.eiger("quickshot", ["", "", "1", "0.1"], dcm, detector_stage)
    )
    assert fake_caput.call_count == 30


@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caput")
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caget")
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.bps.sleep")
def test_eiger_triggered(_, fake_caget, fake_caput, run_engine, dcm, detector_stage):
    set_mock_value(detector_stage.z.user_readback, 300)
    set_mock_value(dcm.wavelength_in_a.user_readback, 1)
    run_engine(
        setup_beamline.eiger("triggered", ["", "", "10", "0.1"], dcm, detector_stage)
    )
    assert fake_caget.call_count == 2
    assert fake_caput.call_count == 30



import json
from pathlib import Path

from bluesky.callbacks import CallbackBase

from mx_bluesky.beamlines.i24.parameters.constants import PlanNameConstants
from mx_bluesky.common.parameters.rotation import SingleRotationScan
from mx_bluesky.common.utils.log import LOGGER

READING_DUMP_FILENAME = "collection_info.json"


class JsonMetadataWriter(CallbackBase):
    """Callback class to handle the creation of metadata json files for commissioning.

    To use, subscribe the Bluesky RunEngine to an instance of this class.
    E.g.:
        metadata_writer_callback = JsonMetadataWriter(parameters)
        RE.subscribe(metadata_writer_callback)
    Or decorate a plan using bluesky.preprocessors.subs_decorator.

    See: https://blueskyproject.io/bluesky/callbacks.html#ways-to-invoke-callbacks

    """

    def __init__(self):
        self.wavelength_in_a = None
        self.energy_in_kev = None
        self.detector_distance_mm = None
        self.final_path: Path | None = None
        self.descriptors: dict[str, dict] = {}
        self.transmission: float | None = None
        self.parameters: SingleRotationScan | None = None

        super().__init__()

    def start(self, doc: dict):  # type: ignore
        if doc.get("subplan_name") == PlanNameConstants.ROTATION_MAIN:
            json_params = doc.get("rotation_scan_params")
            assert json_params is not None
            LOGGER.info(
                f"Metadata writer recieved start document with experiment parameters {json_params}"
            )
            self.parameters = SingleRotationScan(**json.loads(json_params))
            self.run_start_uid = doc.get("uid")

    def descriptor(self, doc: dict):  # type: ignore
        self.descriptors[doc["uid"]] = doc

    def event(self, doc: dict):  # type: ignore
        event_descriptor = self.descriptors[doc["descriptor"]]

        if event_descriptor.get("name") == PlanNameConstants.ROTATION_DEVICE_READ:
            assert self.parameters is not None
            data = doc.get("data")
            assert data is not None
            self.wavelength_in_a = data.get("dcm-wavelength_in_a")
            self.energy_in_kev = data.get("dcm-energy_in_keV")
            self.detector_distance_mm = data.get("detector_motion-z")
            assert data.get("detector-_writer-file_path"), (
                "No detector writer path was found"
            )
            self.final_path = Path(data.get("detector-_writer-file_path"))

            LOGGER.info(
                f"Metadata writer received parameters, energy_in_kev: {self.energy_in_kev}, wavelength: {self.wavelength_in_a}, det_distance_mm: {self.detector_distance_mm}, file path: {self.final_path}"
            )

    def stop(self, doc: dict):  # type: ignore
        assert self.parameters is not None
        if (
            self.run_start_uid is not None
            and doc.get("run_start") == self.run_start_uid
            and self.final_path
        ):
            with open(self.final_path / READING_DUMP_FILENAME, "w") as f:
                f.write(
                    json.dumps(
                        {
                            "wavelength_in_a": self.wavelength_in_a,
                            "energy_kev": self.energy_in_kev,
                            "angular_increment_deg": self.parameters.rotation_increment_deg,
                            "detector_distance_mm": self.detector_distance_mm,
                        }
                    )
                )



from __future__ import annotations

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
from bluesky.utils import MsgGenerator
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.motors import XYZStage
from dodal.devices.robot import SAMPLE_LOCATION_EMPTY, BartRobot
from dodal.devices.smargon import CombinedMove, Smargon, StubPosition
from dodal.plan_stubs.motor_utils import MoveTooLargeError, home_and_reset_wrapper

from mx_bluesky.common.parameters.constants import (
    DocDescriptorNames,
    HardwareConstants,
    PlanNameConstants,
)


def _raise_exception_if_moved_out_of_cryojet(exception):
    yield from bps.null()
    if isinstance(exception, MoveTooLargeError):
        raise Exception(
            f"Moving {exception.axis} back to {exception.position} after \
                        robot load would move it out of the cryojet. The max safe \
                        distance is {exception.maximum_move}"
        )


def do_plan_while_lower_gonio_at_home(plan: MsgGenerator, lower_gonio: XYZStage):
    """Moves the lower gonio to home then performs the provided plan and moves it back.

    The lower gonio must be in the correct position for the robot load and we
    want to put it back afterwards. Note we don't need to wait for the move as the robot
    is interlocked to the lower gonio and the move is quicker than the robot takes to
    get to the load position.

    Args:
        plan (MsgGenerator): The plan to run while the lower gonio is at home.
        lower_gonio (XYZStage): The lower gonio to home.
    """
    yield from bpp.contingency_wrapper(
        home_and_reset_wrapper(
            plan,
            lower_gonio,
            BartRobot.LOAD_TOLERANCE_MM,
            HardwareConstants.CRYOJET_MARGIN_MM,
            "lower_gonio",
            wait_for_all=False,
        ),
        except_plan=_raise_exception_if_moved_out_of_cryojet,
    )
    return "reset-lower_gonio"


def prepare_for_robot_load(
    aperture_scatterguard: ApertureScatterguard, smargon: Smargon
):
    yield from bps.abs_set(
        aperture_scatterguard.selected_aperture,
        ApertureValue.OUT_OF_BEAM,
        group="prepare_robot_load",
    )

    yield from bps.mv(smargon.stub_offsets, StubPosition.RESET_TO_ROBOT_LOAD)

    yield from bps.mv(smargon, CombinedMove(x=0, y=0, z=0, chi=0, phi=0, omega=0))

    yield from bps.wait("prepare_robot_load")


def robot_unload(
    robot: BartRobot,
    smargon: Smargon,
    aperture_scatterguard: ApertureScatterguard,
    lower_gonio: XYZStage,
    visit: str | None,
):
    """Unloads the currently mounted pin into the location that it was loaded from. The
    loaded location is stored on the robot and so need not be provided.
    """
    yield from prepare_for_robot_load(aperture_scatterguard, smargon)
    sample_id = yield from bps.rd(robot.sample_id)

    @bpp.run_decorator(
        md={
            "subplan_name": PlanNameConstants.ROBOT_UNLOAD,
            "metadata": {"visit": visit, "sample_id": sample_id},
            "activate_callbacks": [
                "RobotLoadISPyBCallback",
            ],
        },
    )
    def do_robot_unload_and_send_to_ispyb():
        yield from bps.create(name=DocDescriptorNames.ROBOT_UPDATE)
        yield from bps.read(robot)
        yield from bps.save()

        def _unload():
            yield from bps.abs_set(robot, SAMPLE_LOCATION_EMPTY, wait=True)

        gonio_finished = yield from do_plan_while_lower_gonio_at_home(
            _unload(), lower_gonio
        )
        yield from bps.wait(gonio_finished)

    yield from do_robot_unload_and_send_to_ispyb()



"""Internal parameter models for the private use of hyperion plan runners only."""

from pydantic import BaseModel


class Wait(BaseModel):
    """Represents an instruction from Agamemnon for Hyperion to wait for a specified time
    Attributes:
        duration_s: duration to wait in seconds
    """

    duration_s: float


class UDCDefaultState(BaseModel):
    """Represents an instruction to execute the UDC default state plan."""

    pass


class UDCCleanup(BaseModel):
    """Represents an instruction to perform UDC Cleanup,
    in which the detector shutter is closed."""

    pass


class RobotUnload(BaseModel):
    """Represents an instruction to perform a robot unload."""

    pass



from pydantic.dataclasses import dataclass


@dataclass(frozen=True)
class PlanNameConstants:
    ROTATION_DEVICE_READ = "ROTATION DEVICE READ"
    SINGLE_ROTATION_SCAN = "OUTER SINGLE ROTATION SCAN"
    MULTI_ROTATION_SCAN = "OUTER MULTI ROTATION SCAN"
    ROTATION_MAIN = "ROTATION MAIN"



from .jungfrau_commissioning.experiment_plans.do_darks import (
    do_non_pedestal_darks,
    do_pedestal_darks,
)
from .serial.extruder.i24ssx_extruder_collect_py3v2 import (
    enter_hutch,
    initialise_extruder,
    laser_check,
)
from .serial.fixed_target.i24ssx_chip_manager_py3v1 import (
    block_check,
    cs_maker,
    fiducial,
    initialise_stages,
    moveto,
    moveto_preset,
)
from .web_gui_plans.general_plans import (
    gui_gonio_move_on_click,
    gui_move_backlight,
    gui_move_detector,
    gui_run_chip_collection,
    gui_run_extruder_collection,
    gui_set_fiducial_0,
    gui_set_zoom_level,
    gui_stage_move_on_click,
)
from .web_gui_plans.jungfrau_plans import gui_run_jf_rotation_scan
from .web_gui_plans.oav_plans import (
    focus_on_oav_view,
    move_block_on_arrow_click,
    move_nudge_on_arrow_click,
    move_window_on_arrow_click,
)

# NOTE. Only plans that will be used by cluster blueapi/web UI

__all__ = [
    "gui_stage_move_on_click",
    "gui_gonio_move_on_click",
    "gui_move_detector",
    "gui_run_chip_collection",
    "gui_move_backlight",
    "gui_set_zoom_level",
    "gui_set_fiducial_0",
    "gui_run_extruder_collection",
    "initialise_extruder",
    "enter_hutch",
    "laser_check",
    "moveto",
    "moveto_preset",
    "block_check",
    "cs_maker",
    "fiducial",
    "initialise_stages",
    "focus_on_oav_view",
    "move_block_on_arrow_click",
    "move_nudge_on_arrow_click",
    "move_window_on_arrow_click",
    # Jungfrau specific
    "do_pedestal_darks",
    "do_non_pedestal_darks",
    "gui_run_jf_rotation_scan",
]



from unittest.mock import MagicMock, Mock, patch

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
import pytest
from bluesky.run_engine import RunEngine
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.robot import BartRobot
from dodal.devices.webcam import Webcam
from ophyd_async.core import set_mock_value

from mx_bluesky.common.external_interaction.ispyb.exp_eye_store import BLSampleStatus
from mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback import (
    RobotLoadISPyBCallback,
)
from mx_bluesky.hyperion.parameters.constants import CONST

VISIT = "cm31105-4"

SAMPLE_ID = 231412
SAMPLE_PUCK = 50
SAMPLE_PIN = 4
ACTION_ID = 1098

metadata = {
    "subplan_name": CONST.PLAN.ROBOT_LOAD,
    "metadata": {
        "visit": VISIT,
        "sample_id": SAMPLE_ID,
        "sample_puck": SAMPLE_PUCK,
        "sample_pin": SAMPLE_PIN,
    },
    "activate_callbacks": [
        "RobotLoadISPyBCallback",
    ],
}

metadata_unload_no_visit = {
    "subplan_name": CONST.PLAN.ROBOT_UNLOAD,
    "metadata": {
        "sample_puck": SAMPLE_PUCK,
        "sample_pin": SAMPLE_PIN,
    },
    "activate_callbacks": [
        "RobotLoadISPyBCallback",
    ],
}

update_doc_data = {
    "sampleBarcode": "BARCODE",
    "xtalSnapshotBefore": "test_webcam_snapshot",
    "xtalSnapshotAfter": "test_oav_snapshot",
    "containerLocation": SAMPLE_PIN,
    "dewarLocation": SAMPLE_PUCK,
}


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback.ExpeyeInteraction",
    autospec=True,
)
def test_given_start_doc_with_expected_data_then_data_put_in_ispyb(
    expeye: MagicMock, run_engine: RunEngine
):
    run_engine.subscribe(RobotLoadISPyBCallback())
    expeye.return_value.start_robot_action.return_value = ACTION_ID

    @bpp.run_decorator(md=metadata)
    def my_plan():
        yield from bps.null()

    run_engine(my_plan())

    expeye.return_value.start_robot_action.assert_called_once_with(
        "LOAD", "cm31105", 4, 231412
    )
    expeye.return_value.end_robot_action.assert_called_once_with(
        ACTION_ID, "success", "OK"
    )


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback.ExpeyeInteraction",
    autospec=True,
)
def test_given_failing_plan_then_exception_detail(
    expeye: MagicMock, run_engine: RunEngine
):
    run_engine.subscribe(RobotLoadISPyBCallback())
    expeye.return_value.start_robot_action.return_value = ACTION_ID

    class _Error(Exception): ...

    @bpp.run_decorator(md=metadata)
    def my_plan():
        raise _Error("BAD")
        yield from bps.null()

    with pytest.raises(_Error):
        run_engine(my_plan())

    expeye.return_value.start_robot_action.assert_called_once_with(
        "LOAD", "cm31105", 4, 231412
    )
    expeye.return_value.end_robot_action.assert_called_once_with(
        ACTION_ID, "fail", "BAD"
    )


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback.ExpeyeInteraction.end_robot_action"
)
@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback.ExpeyeInteraction"
    ".update_sample_status",
    new=MagicMock(),
)
def test_given_end_called_but_no_start_then_exception_raised(end_load):
    callback = RobotLoadISPyBCallback()
    callback.active = True
    with pytest.raises(AssertionError):
        callback.activity_gated_stop({"run_uid": None})  # type: ignore
    end_load.assert_not_called()


@bpp.run_decorator(md=metadata)
def successful_robot_load_plan(robot: BartRobot, oav: OAV, webcam: Webcam):
    yield from bps.create(name=CONST.DESCRIPTORS.ROBOT_UPDATE)
    yield from bps.read(robot)
    yield from bps.read(oav.snapshot)
    yield from bps.read(webcam)
    yield from bps.save()


@bpp.run_decorator(md=metadata)
def unsuccessful_robot_load_plan():
    yield from []
    raise AssertionError("Test failure")


@bpp.run_decorator(md=metadata_unload_no_visit)
def robot_unload_plan(robot: BartRobot, oav: OAV, webcam: Webcam):
    yield from bps.create(name=CONST.DESCRIPTORS.ROBOT_UPDATE)
    yield from bps.read(robot)
    yield from bps.read(oav.snapshot)
    yield from bps.read(webcam)
    yield from bps.save()


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback.ExpeyeInteraction",
    autospec=True,
)
def test_given_plan_reads_robot_then_data_put_in_ispyb(
    expeye: MagicMock, robot: BartRobot, oav: OAV, webcam: Webcam, run_engine: RunEngine
):
    run_engine.subscribe(RobotLoadISPyBCallback())
    expeye.return_value.start_robot_action.return_value = ACTION_ID

    set_mock_value(oav.snapshot.last_saved_path, "test_oav_snapshot")
    set_mock_value(webcam.last_saved_path, "test_webcam_snapshot")
    set_mock_value(robot.sample_id, SAMPLE_ID)
    set_mock_value(robot.current_pin, SAMPLE_PIN)
    set_mock_value(robot.current_puck, SAMPLE_PUCK)

    run_engine(successful_robot_load_plan(robot, oav, webcam))

    expeye.return_value.start_robot_action.assert_called_once_with(
        "LOAD", "cm31105", 4, 231412
    )
    expeye.return_value.update_robot_action.assert_called_once_with(
        ACTION_ID, update_doc_data
    )
    expeye.return_value.end_robot_action.assert_called_once_with(
        ACTION_ID, "success", "OK"
    )


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback.ExpeyeInteraction",
    autospec=True,
)
def test_robot_load_complete_triggers_bl_sample_status_loaded(
    mock_sample_handling,
    run_engine: RunEngine,
    robot: BartRobot,
    oav: OAV,
    webcam: Webcam,
):
    run_engine.subscribe(RobotLoadISPyBCallback())

    run_engine(successful_robot_load_plan(robot, oav, webcam))

    mock_sample_handling.return_value.update_sample_status.assert_called_with(
        SAMPLE_ID, BLSampleStatus.LOADED
    )


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback.ExpeyeInteraction",
    autospec=True,
)
def test_robot_load_fails_triggers_bl_sample_status_error(
    mock_sample_handling,
    run_engine: RunEngine,
    robot: BartRobot,
    oav: OAV,
    webcam: Webcam,
):
    run_engine.subscribe(RobotLoadISPyBCallback())

    with pytest.raises(AssertionError, match="Test failure"):
        run_engine(unsuccessful_robot_load_plan())

    mock_sample_handling.return_value.update_sample_status.assert_called_with(
        SAMPLE_ID, BLSampleStatus.ERROR_BEAMLINE
    )


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback.ExpeyeInteraction",
    autospec=True,
)
def test_robot_unload_event_without_sample_id_and_visit_is_ignored(
    mock_expeye: MagicMock,
    run_engine: RunEngine,
    robot: BartRobot,
    oav: OAV,
    webcam: Webcam,
):
    unwrapped = RobotLoadISPyBCallback()
    callback = Mock(wraps=unwrapped)
    run_engine.subscribe(callback)

    run_engine(robot_unload_plan(robot, oav, webcam))
    mock_expeye.return_value.start_robot_action.assert_not_called()
    mock_expeye.return_value.update_robot_action.assert_not_called()
    mock_expeye.return_value.end_robot_action.assert_not_called()
    mock_expeye.return_value.update_sample_status.assert_not_called()


# In the event where sample id has not been set on the robot - this
# condition was encountered during testing although it's not clear how
# the system arrived in this state.
@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback.ExpeyeInteraction",
    autospec=True,
)
def test_robot_unload_event_where_sample_id_is_zero_is_ignored(
    expeye_interaction: MagicMock,
    test_event_data,
):
    ispyb_callback = RobotLoadISPyBCallback()
    expeye = expeye_interaction.return_value
    expeye.start_robot_action.return_value = ACTION_ID
    start_document = test_event_data.test_robot_unload_start_document.copy()
    start_document["metadata"]["sample_id"] = 0
    ispyb_callback.start(start_document)
    ispyb_callback.descriptor(test_event_data.test_descriptor_document_robot_unload)
    ispyb_callback.event(test_event_data.test_event_document_robot_unload)
    ispyb_callback.stop(test_event_data.test_robot_unload_stop_document)

    expeye.start_robot_action.assert_not_called()
    expeye.update_robot_action.assert_not_called()
    expeye.end_robot_action.assert_not_called()
    expeye.update_sample_status.assert_not_called()


# When udc default state unloads the robot.
@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback.ExpeyeInteraction",
    autospec=True,
)
def test_robot_unload_event_where_visit_is_undefined_is_ignored(
    expeye_interaction: MagicMock,
    test_event_data,
):
    ispyb_callback = RobotLoadISPyBCallback()
    expeye = expeye_interaction.return_value
    expeye.start_robot_action.return_value = ACTION_ID
    start_document = test_event_data.test_robot_unload_start_document.copy()
    del start_document["metadata"]["visit"]
    ispyb_callback.start(start_document)
    ispyb_callback.descriptor(test_event_data.test_descriptor_document_robot_unload)
    ispyb_callback.event(test_event_data.test_event_document_robot_unload)
    ispyb_callback.stop(test_event_data.test_robot_unload_stop_document)

    expeye.start_robot_action.assert_not_called()
    expeye.update_robot_action.assert_not_called()
    expeye.end_robot_action.assert_not_called()
    expeye.update_sample_status.assert_not_called()



from collections.abc import Sequence

from blueapi.client.client import BlueapiClient
from blueapi.client.event_bus import BlueskyStreamingError
from blueapi.config import ApplicationConfig
from blueapi.core import BlueskyContext
from blueapi.service.model import TaskRequest
from blueapi.worker.event import TaskError
from bluesky import plan_stubs as bps
from bluesky.utils import MsgGenerator
from pydantic import BaseModel

from mx_bluesky.common.parameters.constants import Status
from mx_bluesky.common.utils.exceptions import CrystalNotFoundError, SampleError
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion._plan_runner_params import (
    RobotUnload,
    UDCCleanup,
    UDCDefaultState,
    Wait,
)
from mx_bluesky.hyperion.blueapi.parameters import LoadCentreCollectParams
from mx_bluesky.hyperion.plan_runner import PlanError, PlanRunner
from mx_bluesky.hyperion.supervisor._task_monitor import TaskMonitor


class SupervisorRunner(PlanRunner):
    """Runner that executes plans by delegating to a remote blueapi instance"""

    def __init__(
        self,
        bluesky_context: BlueskyContext,
        client_config: ApplicationConfig,
        dev_mode: bool,
    ):
        super().__init__(bluesky_context, dev_mode)
        self.blueapi_client = BlueapiClient.from_config(client_config)
        self._current_status = Status.IDLE

    def decode_and_execute(
        self, current_visit: str | None, parameter_list: Sequence[BaseModel]
    ) -> MsgGenerator:
        try:
            yield from self.check_external_callbacks_are_alive()
        except Exception as e:
            raise PlanError(f"Exception raised during plan execution: {e}") from e
        instrument_session = current_visit or "NO_VISIT"
        try:
            if self._current_status == Status.ABORTING:
                raise PlanError("Plan execution cancelled, supervisor is shutting down")
            self._current_status = Status.BUSY
            try:
                for parameters in parameter_list:
                    LOGGER.info(
                        f"Executing plan with parameters: {parameters.model_dump_json(indent=2)}"
                    )
                    match parameters:
                        case LoadCentreCollectParams():
                            task_request = TaskRequest(
                                name="load_centre_collect",
                                params={"parameters": parameters},
                                instrument_session=instrument_session,
                            )
                            self._run_task_remotely(task_request)
                        case RobotUnload():
                            task_request = TaskRequest(
                                name="robot_unload",
                                params={"visit": current_visit},
                                instrument_session=instrument_session,
                            )
                            self._run_task_remotely(task_request)
                        case Wait():
                            yield from bps.sleep(parameters.duration_s)
                        case UDCDefaultState():
                            task_request = TaskRequest(
                                name="move_to_udc_default_state",
                                params={},
                                instrument_session=instrument_session,
                            )
                            self._run_task_remotely(task_request)
                        case UDCCleanup():
                            task_request = TaskRequest(
                                name="clean_up_udc",
                                params={},
                                instrument_session=instrument_session,
                            )
                            self._run_task_remotely(task_request)
                        case _:
                            raise AssertionError(
                                f"Unsupported instruction decoded from agamemnon {type(parameters)}"
                            )
            except SampleError:
                LOGGER.info("Ignoring sample error, continuing...")
        except:
            self._current_status = Status.FAILED
            raise
        else:
            self._current_status = Status.IDLE
        return current_visit

    @property
    def current_status(self) -> Status:
        return self._current_status

    def is_connected(self) -> bool:
        try:
            self.blueapi_client.state  # noqa: B018
        except Exception as e:
            LOGGER.debug(f"Failed to get worker state: {e}")
            return False
        return True

    def shutdown(self):
        LOGGER.info(
            "Hyperion supervisor received shutdown request, signalling abort to BlueAPI server..."
        )
        if self.current_status != Status.BUSY:
            self.request_run_engine_abort()
        else:
            self._current_status = Status.ABORTING
            self.blueapi_client.abort()

    def _run_task_remotely(self, task_request: TaskRequest):
        try:
            with TaskMonitor(self.blueapi_client, task_request) as task_monitor:
                task_status = self.blueapi_client.run_task(
                    task_request, on_event=task_monitor.on_blueapi_event
                )
            LOGGER.info(
                f"hyperion-blueapi completed task execution with task_status {task_status}"
            )

            match task_status.result:
                case TaskError() as task_error:
                    LOGGER.info(
                        f"hyperion-blueapi plan execution encountered an error: {task_error.type}: {task_error.message}"
                    )
                    match task_error.type:
                        case CrystalNotFoundError.__name__ | SampleError.__name__:
                            raise SampleError(
                                f"Remote task error: {task_error.message}"
                            )
                        case _:
                            if self.current_status != Status.ABORTING:
                                raise PlanError(
                                    f"Exception raised during plan execution: {task_error}"
                                )
        except BlueskyStreamingError as e:
            if self.current_status != Status.ABORTING:
                raise PlanError(
                    "BlueskyStreamingError raised during plan execution"
                ) from e
        finally:
            if self.current_status == Status.ABORTING:
                LOGGER.info("Aborting local runner...")
                self.request_run_engine_abort()



from collections.abc import Callable
from unittest.mock import ANY, MagicMock

import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from bluesky.utils import Msg
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.motors import XYZStage
from dodal.devices.robot import SAMPLE_LOCATION_EMPTY, BartRobot
from dodal.devices.smargon import CombinedMove, Smargon, StubPosition
from ophyd_async.core import completed_status, get_mock_put, set_mock_value

from mx_bluesky.common.device_setup_plans.robot_load_unload import (
    prepare_for_robot_load,
    robot_unload,
)
from mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback import (
    RobotLoadISPyBCallback,
)

TEST_PUCK = 1
TEST_PIN = 2
TEST_SAMPLE_ID = 123456


@pytest.fixture
def loaded_robot(robot: BartRobot):
    set_mock_value(robot.current_pin, TEST_PIN)
    set_mock_value(robot.current_puck, TEST_PUCK)
    set_mock_value(robot.sample_id, TEST_SAMPLE_ID)
    return robot


# Remove when in bluesky proper, see https://github.com/bluesky/bluesky/issues/1924
def assert_messages_any_order(messages: list, predicates: list[Callable[[Msg], bool]]):
    remaining_predicates = set(predicates)

    for message in messages:
        matched = {p for p in remaining_predicates if p(message)}
        remaining_predicates -= matched

        if not remaining_predicates:
            break

    assert not remaining_predicates, f"Unmatched predicates: {remaining_predicates}"
    return messages


async def test_when_prepare_for_robot_load_called_then_moves_as_expected(
    aperture_scatterguard: ApertureScatterguard,
    smargon: Smargon,
    run_engine: RunEngine,
):
    smargon.stub_offsets.set = MagicMock(side_effect=lambda _: completed_status())
    get_mock_put(aperture_scatterguard.selected_aperture).reset_mock()

    set_mock_value(smargon.x.user_setpoint, 10)
    set_mock_value(smargon.z.user_setpoint, 5)
    set_mock_value(smargon.omega.user_setpoint, 90)

    run_engine(prepare_for_robot_load(aperture_scatterguard, smargon))

    assert await smargon.x.user_setpoint.get_value() == 0
    assert await smargon.z.user_setpoint.get_value() == 0
    assert await smargon.omega.user_setpoint.get_value() == 0

    smargon.stub_offsets.set.assert_called_once_with(StubPosition.RESET_TO_ROBOT_LOAD)  # type: ignore
    get_mock_put(aperture_scatterguard.selected_aperture).assert_called_once_with(
        ApertureValue.OUT_OF_BEAM
    )


async def test_when_robot_unload_called_then_sample_area_prepared_before_load(
    robot: BartRobot,
    smargon: Smargon,
    aperture_scatterguard: ApertureScatterguard,
    lower_gonio: XYZStage,
    sim_run_engine: RunEngineSimulator,
):
    msgs = sim_run_engine.simulate_plan(
        robot_unload(robot, smargon, aperture_scatterguard, lower_gonio, "")
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == aperture_scatterguard.selected_aperture.name
        and msg.args[0] == ApertureValue.OUT_OF_BEAM,
    )

    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "gonio"
        and msg.args[0] == CombinedMove(x=0, y=0, z=0, omega=0, chi=0, phi=0),
    )

    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is robot
        and msg.args[0] == SAMPLE_LOCATION_EMPTY,
    )
    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait"
        and msg.kwargs.get("group") == msgs[0].kwargs["group"],
    )


async def test_given_lower_gonio_needs_moving_then_it_is_homed_before_unload_and_put_back_after(
    robot: BartRobot,
    smargon: Smargon,
    aperture_scatterguard: ApertureScatterguard,
    lower_gonio: XYZStage,
    sim_run_engine: RunEngineSimulator,
):
    # Replace when https://github.com/bluesky/bluesky/issues/1906 is fixed
    def locate_gonio(_):
        return {"readback": 0.1}

    sim_run_engine.add_handler("locate", locate_gonio, lower_gonio.x.name)
    sim_run_engine.add_handler("locate", locate_gonio, lower_gonio.y.name)
    sim_run_engine.add_handler("locate", locate_gonio, lower_gonio.z.name)

    sim_run_engine.add_read_handler_for(robot.sample_id, 1000)

    msgs = sim_run_engine.simulate_plan(
        robot_unload(robot, smargon, aperture_scatterguard, lower_gonio, "")
    )

    msgs = assert_messages_any_order(
        msgs,
        [
            lambda msg: msg.command == "set"
            and msg.obj.name == f"lower_gonio-{axis}"
            and msg.args[0] == 0
            for axis in ["x", "y", "z"]
        ],
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is robot
        and msg.args[0] == SAMPLE_LOCATION_EMPTY,
    )

    msgs = assert_messages_any_order(
        msgs,
        [
            lambda msg: msg.command == "set"
            and msg.obj.name == f"lower_gonio-{axis}"
            and msg.args[0] == 0.1
            for axis in ["x", "y", "z"]
        ],
    )


def test_when_unload_plan_run_then_initial_unload_ispyb_deposition_made(
    run_engine: RunEngine,
    robot: BartRobot,
    smargon: Smargon,
    aperture_scatterguard: ApertureScatterguard,
    lower_gonio: XYZStage,
):
    callback = RobotLoadISPyBCallback()
    callback.expeye = (mock_expeye := MagicMock())
    run_engine.subscribe(callback)

    set_mock_value(robot.sample_id, expected_sample_id := 1234)

    run_engine(
        robot_unload(robot, smargon, aperture_scatterguard, lower_gonio, "cm37235-2")
    )

    mock_expeye.start_robot_action.assert_called_once_with(
        "UNLOAD", "cm37235", 2, expected_sample_id
    )


def test_when_unload_plan_run_then_full_ispyb_deposition_made(
    run_engine: RunEngine,
    robot: BartRobot,
    smargon: Smargon,
    aperture_scatterguard: ApertureScatterguard,
    lower_gonio: XYZStage,
):
    callback = RobotLoadISPyBCallback()
    callback.expeye = (mock_expeye := MagicMock())
    run_engine.subscribe(callback)

    set_mock_value(robot.sample_id, expected_sample_id := 1234)
    set_mock_value(robot.current_pin, expected_pin := 12)
    set_mock_value(robot.current_puck, expected_puck := 45)
    set_mock_value(robot.barcode, expected_barcode := "BARCODE")

    action_id = 1098
    mock_expeye.start_robot_action.return_value = action_id

    run_engine(
        robot_unload(robot, smargon, aperture_scatterguard, lower_gonio, "cm37235-2")
    )

    mock_expeye.start_robot_action.assert_called_once_with(
        "UNLOAD", "cm37235", 2, expected_sample_id
    )
    mock_expeye.update_robot_action.assert_called_once_with(
        action_id,
        {
            "sampleBarcode": expected_barcode,
            "containerLocation": expected_pin,
            "dewarLocation": expected_puck,
        },
    )
    mock_expeye.end_robot_action.assert_called_once_with(action_id, "success", "OK")


def test_when_unload_plan_fails_then_error_deposited_in_ispyb(
    run_engine: RunEngine,
    loaded_robot: BartRobot,
    smargon: Smargon,
    aperture_scatterguard: ApertureScatterguard,
    lower_gonio: XYZStage,
):
    class TestError(Exception): ...

    callback = RobotLoadISPyBCallback()
    callback.expeye = (mock_expeye := MagicMock())
    run_engine.subscribe(callback)
    loaded_robot.set = MagicMock(side_effect=TestError("Bad Error"))

    action_id = 1098
    mock_expeye.start_robot_action.return_value = action_id

    with pytest.raises(TestError):
        run_engine(
            robot_unload(
                loaded_robot, smargon, aperture_scatterguard, lower_gonio, "cm37235-2"
            )
        )

    mock_expeye.start_robot_action.assert_called_once_with("UNLOAD", "cm37235", 2, ANY)
    mock_expeye.end_robot_action.assert_called_once_with(action_id, "fail", "Bad Error")



import asyncio
import pprint
import sys
from collections.abc import Generator
from functools import partial
from pathlib import Path
from typing import Any, cast
from unittest.mock import MagicMock, patch

import pytest
from _pytest.fixtures import FixtureRequest
from bluesky.run_engine import RunEngine
from dodal.beamlines import i03
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.backlight import Backlight
from dodal.devices.beamlines.i24.commissioning_jungfrau import CommissioningJungfrau
from dodal.devices.beamsize.beamsize import BeamsizeBase
from dodal.devices.detector.detector_motion import DetectorMotion
from dodal.devices.eiger import EigerDetector
from dodal.devices.fast_grid_scan import PandAFastGridScan, ZebraFastGridScanThreeD
from dodal.devices.flux import Flux
from dodal.devices.hutch_shutter import ShutterState
from dodal.devices.mx_phase1.beamstop import Beamstop
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.robot import BartRobot
from dodal.devices.s4_slit_gaps import S4SlitGaps
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import Synchrotron, SynchrotronMode
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutterState
from dodal.devices.zocalo import ZocaloResults
from event_model import RunStart
from event_model.documents import Event
from ophyd_async.core import (
    AsyncStatus,
    AutoMaxIncrementingPathProvider,
    PathInfo,
    PathProvider,
    completed_status,
    init_devices,
    set_mock_value,
)
from ophyd_async.fastcs.panda import HDFPanda

from mx_bluesky.common.experiment_plans.beamstop_check import BeamstopCheckDevices
from mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan import (
    BeamlineSpecificFGSFeatures,
)
from mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback import (
    ZocaloCallback,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanISPyBCallback,
    generate_start_info_from_omega_map,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback import (
    GridscanNexusFileCallback,
)
from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGroupInfo,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)
from mx_bluesky.common.parameters.constants import (
    DocDescriptorNames,
    EnvironmentConstants,
    PlanNameConstants,
)
from mx_bluesky.common.parameters.device_composites import (
    FlyScanEssentialDevices,
    GridDetectThenXRayCentreComposite,
)
from mx_bluesky.common.parameters.gridscan import GridCommon, SpecifiedThreeDGridScan
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionGridDetectThenXRayCentreComposite,
)
from tests.conftest import raw_params_from_file

_ALLOWED_PYTEST_TASKS = {"async_finalizer", "async_setup", "async_teardown"}


def _error_and_kill_pending_tasks(
    loop: asyncio.AbstractEventLoop, test_name: str, test_passed: bool
) -> set[asyncio.Task]:
    """Cancels pending tasks in the event loop for a test. Raises an exception if
    the test hasn't already.

    Args:
        loop: The event loop to check for pending tasks.
        test_name: The name of the test.
        test_passed: Indicates whether the test passed.

    Returns:
        set[asyncio.Task]: The set of unfinished tasks that were cancelled.

    Raises:
        RuntimeError: If there are unfinished tasks and the test didn't fail.
    """
    unfinished_tasks = {
        task
        for task in asyncio.all_tasks(loop)
        if hasattr(coro := task.get_coro(), "__name__")
        and coro.__name__ not in _ALLOWED_PYTEST_TASKS  # type: ignore
        and not task.done()
    }
    for task in unfinished_tasks:
        task.cancel()

    # We only raise an exception here if the test didn't fail anyway.
    # If it did then it makes sense that there's some tasks we need to cancel,
    # but an exception will already have been raised.
    if unfinished_tasks and test_passed:
        raise RuntimeError(
            f"Not all tasks closed during test {test_name}:\n"
            f"{pprint.pformat(unfinished_tasks, width=88)}"
        )

    return unfinished_tasks


@pytest.fixture(autouse=True, scope="function")
async def fail_test_on_unclosed_tasks(request: FixtureRequest):
    """
    Used on every test to ensure failure if there are pending tasks
    by the end of the test.
    """

    fail_count = request.session.testsfailed
    loop = asyncio.get_running_loop()

    loop.set_debug(True)

    request.addfinalizer(
        lambda: _error_and_kill_pending_tasks(
            loop, request.node.name, request.session.testsfailed == fail_count
        )
    )


BASIC_PRE_SETUP_DOC = {
    "undulator-current_gap": 0,
    "synchrotron-synchrotron_mode": SynchrotronMode.USER,
    "s4_slit_gaps-xgap": 0,
    "s4_slit_gaps-ygap": 0,
    "gonio-x": 10.0,
    "gonio-y": 20.0,
    "gonio-z": 30.0,
}

BASIC_POST_SETUP_DOC = {
    "aperture_scatterguard-selected_aperture": ApertureValue.OUT_OF_BEAM,
    "attenuator-actual_transmission": 0,
    "flux-flux_reading": 10,
    "dcm-energy_in_keV": 11.105,
    "beamsize-x_um": 50.0,
    "beamsize-y_um": 20.0,
    "eiger_cam_roi_mode": True,
    "eiger-ispyb_detector_id": 78,
}


def assert_event(mock_call, expected):
    actual = mock_call.args[0]
    if "data" in actual:
        actual = actual["data"]
    for k, v in expected.items():
        assert actual[k] == v, f"Mismatch in key {k}, {actual} <=> {expected}"


def create_gridscan_callbacks() -> tuple[
    GridscanNexusFileCallback, GridscanISPyBCallback
]:
    return (
        GridscanNexusFileCallback(param_type=SpecifiedThreeDGridScan),
        GridscanISPyBCallback(
            param_type=SpecifiedThreeDGridScan,
            emit=ZocaloCallback(
                PlanNameConstants.DO_FGS,
                EnvironmentConstants.ZOCALO_ENV,
                generate_start_info_from_omega_map,
            ),
        ),
    )


@pytest.fixture
def use_beamline_t01():
    """Beamline t01 is a beamline for unit tests that just contains a baton, so that
    loading the beamline context does not require importing lots of modules and instantiating
    many devices"""
    with patch.dict("os.environ", {"BEAMLINE": "t01"}):
        import tests.unit_tests.t01

        with (
            patch.dict(sys.modules, {"dodal.beamlines.t01": tests.unit_tests.t01}),
            patch("mx_bluesky.hyperion.in_process_runner.move_to_udc_default_state"),
            patch(
                "mx_bluesky.hyperion.in_process_runner.device_composite_from_context"
            ),
        ):
            yield tests.unit_tests.t01


@pytest.fixture
def mock_subscriptions(test_fgs_params):
    with (
        patch(
            "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
        ),
        patch(
            "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.StoreInIspyb"
        ) as mock_store_in_ispyb,
    ):
        mock_store_in_ispyb.return_value.begin_deposition.return_value = IspybIds(
            data_collection_ids=(100, 200), data_collection_group_id=0
        )
        mock_store_in_ispyb.return_value.update_deposition.return_value = IspybIds(
            data_collection_ids=(100, 200),
            data_collection_group_id=0,
            grid_ids=(0, 0),
        )
        nexus_callback, ispyb_callback = create_gridscan_callbacks()
        ispyb_callback.ispyb = MagicMock(spec=StoreInIspyb)

        yield (nexus_callback, ispyb_callback)


ReWithSubs = tuple[RunEngine, tuple[GridscanNexusFileCallback | GridscanISPyBCallback]]


@pytest.fixture
def run_engine_with_subs(
    run_engine: RunEngine,
    mock_subscriptions: tuple[GridscanNexusFileCallback | GridscanISPyBCallback],
) -> Generator[ReWithSubs, Any, None]:
    for cb in list(mock_subscriptions):
        run_engine.subscribe(cb)
    yield run_engine, mock_subscriptions


@pytest.fixture
def test_fgs_params(tmp_path, patch_beamline_env_variable):
    return SpecifiedThreeDGridScan(
        **raw_params_from_file(
            "tests/test_data/parameter_json_files/good_test_parameters.json", tmp_path
        )
    )


def mock_zocalo_trigger(zocalo: ZocaloResults, result):
    @AsyncStatus.wrap
    async def mock_complete(results):
        await zocalo._put_results(results, {"dcid": 0, "dcgid": 0})

    zocalo.trigger = MagicMock(side_effect=partial(mock_complete, result))


def modified_store_grid_scan_mock(*args, dcids=(0, 0), dcgid=0, **kwargs):
    mock = MagicMock(spec=StoreInIspyb)
    mock.begin_deposition.return_value = IspybIds(
        data_collection_ids=dcids, data_collection_group_id=dcgid
    )
    mock.update_deposition.return_value = IspybIds(
        data_collection_ids=dcids, data_collection_group_id=dcgid, grid_ids=(0, 0)
    )
    return mock


def make_event_doc(data, descriptor="abc123") -> Event:
    return {
        "time": 0,
        "timestamps": {"a": 0},
        "seq_num": 0,
        "uid": "not so random uid",
        "descriptor": descriptor,
        "data": data,
    }


def run_generic_ispyb_handler_setup(
    ispyb_handler: GridscanISPyBCallback,
    params: SpecifiedThreeDGridScan,
):
    """This is useful when testing 'run_gridscan_and_move(...)' because this stuff
    happens at the start of the outer plan."""

    ispyb_handler.active = True
    ispyb_handler.activity_gated_start(
        {
            "subplan_name": PlanNameConstants.GRIDSCAN_OUTER,
            "mx_bluesky_parameters": params.model_dump_json(),
        }  # type: ignore
    )
    ispyb_handler.activity_gated_descriptor(
        {"uid": "123abc", "name": DocDescriptorNames.HARDWARE_READ_PRE}  # type: ignore
    )
    ispyb_handler.activity_gated_event(
        make_event_doc(
            BASIC_PRE_SETUP_DOC,
            descriptor="123abc",
        )
    )
    ispyb_handler.activity_gated_descriptor(
        {"uid": "abc123", "name": DocDescriptorNames.HARDWARE_READ_DURING}  # type: ignore
    )
    ispyb_handler.activity_gated_event(
        make_event_doc(
            BASIC_POST_SETUP_DOC,
            descriptor="abc123",
        )
    )


@pytest.fixture
async def zebra_fast_grid_scan():
    zebra_fast_grid_scan = i03.zebra_fast_grid_scan.build(
        connect_immediately=True, mock=True
    )
    set_mock_value(zebra_fast_grid_scan.device_scan_invalid, 0.0)
    set_mock_value(zebra_fast_grid_scan.x_scan_valid, 1.0)
    set_mock_value(zebra_fast_grid_scan.y_scan_valid, 1.0)
    set_mock_value(zebra_fast_grid_scan.z_scan_valid, 1.0)
    set_mock_value(zebra_fast_grid_scan.position_counter, 0)
    return zebra_fast_grid_scan


@pytest.fixture
async def fake_fgs_composite(
    smargon: Smargon,
    test_fgs_params: SpecifiedThreeDGridScan,
    attenuator,
    xbpm_feedback,
    synchrotron,
    aperture_scatterguard,
    zocalo,
    panda,
    backlight,
):
    fake_composite = FlyScanEssentialDevices(
        # We don't use the eiger fixture here because .unstage() is used in some tests
        eiger=i03.eiger.build(mock=True),
        gonio=smargon,
        synchrotron=synchrotron,
    )

    fake_composite.eiger.stage = MagicMock(side_effect=lambda: completed_status())
    # unstage should be mocked on a per-test basis because several rely on unstage
    fake_composite.eiger.set_detector_parameters(test_fgs_params.detector_params)
    fake_composite.eiger.stop_odin_when_all_frames_collected = MagicMock()
    fake_composite.eiger.odin.check_and_wait_for_odin_state = lambda timeout: True

    test_result = {
        "centre_of_mass": [6, 6, 6],
        "max_voxel": [5, 5, 5],
        "max_count": 123456,
        "n_voxels": 321,
        "total_count": 999999,
        "bounding_box": [[3, 3, 3], [9, 9, 9]],
        "sample_id": 12345,
    }

    @AsyncStatus.wrap
    async def mock_complete(result):
        await zocalo._put_results([result], {"dcid": 0, "dcgid": 0})

    zocalo.trigger = MagicMock(side_effect=partial(mock_complete, test_result))  # type: ignore
    zocalo.timeout_s = 3
    set_mock_value(fake_composite.gonio.x.max_velocity, 10)

    return fake_composite


@pytest.fixture
def dummy_rotation_data_collection_group_info():
    return DataCollectionGroupInfo(
        visit_string="cm31105-4",
        experiment_type="SAD",
        sample_id=364758,
    )


@pytest.fixture
def beamline_specific(
    zebra_fast_grid_scan: ZebraFastGridScanThreeD,
) -> BeamlineSpecificFGSFeatures:
    return BeamlineSpecificFGSFeatures(
        setup_trigger_plan=MagicMock(),
        tidy_plan=MagicMock(),
        set_flyscan_params_plan=MagicMock(),
        fgs_motors=zebra_fast_grid_scan,
        read_pre_flyscan_plan=MagicMock(),
        read_during_collection_plan=MagicMock(),
    )


@pytest.fixture
def test_full_grid_scan_params(tmp_path):
    params = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_grid_with_edge_detect_parameters.json",
        tmp_path,
    )
    return GridCommon(**params)


@pytest.fixture
async def grid_detect_xrc_devices(
    aperture_scatterguard: ApertureScatterguard,
    backlight: Backlight,
    beamstop_phase1: Beamstop,
    beamsize: BeamsizeBase,
    detector_motion: DetectorMotion,
    eiger: EigerDetector,
    smargon: Smargon,
    oav: OAV,
    ophyd_pin_tip_detection: PinTipDetection,
    zocalo: ZocaloResults,
    synchrotron: Synchrotron,
    fast_grid_scan: ZebraFastGridScanThreeD,
    s4_slit_gaps: S4SlitGaps,
    flux: Flux,
    zebra,
    zebra_shutter,
    xbpm_feedback,
    attenuator,
    undulator,
    dcm,
):
    yield GridDetectThenXRayCentreComposite(
        aperture_scatterguard=aperture_scatterguard,
        attenuator=attenuator,
        backlight=backlight,
        beamstop=beamstop_phase1,
        beamsize=beamsize,
        detector_motion=detector_motion,
        eiger=eiger,
        zebra_fast_grid_scan=fast_grid_scan,
        flux=flux,
        oav=oav,
        pin_tip_detection=ophyd_pin_tip_detection,
        gonio=smargon,
        synchrotron=synchrotron,
        s4_slit_gaps=s4_slit_gaps,
        undulator=undulator,
        xbpm_feedback=xbpm_feedback,
        zebra=zebra,
        zocalo=zocalo,
        dcm=dcm,
        robot=MagicMock(spec=BartRobot),
        sample_shutter=zebra_shutter,
    )


@pytest.fixture
async def hyperion_grid_detect_xrc_devices(grid_detect_xrc_devices):
    composite = cast(HyperionGridDetectThenXRayCentreComposite, grid_detect_xrc_devices)
    composite.panda = MagicMock(spec=HDFPanda)
    composite.panda_fast_grid_scan = MagicMock(spec=PandAFastGridScan)
    return composite


class _BasePathProvider(PathProvider):
    """Minimal adaption of BlueAPI's PathProvider.

    To use this PathProvider in a test, a run must be open
    with 'detector_file_template' attached as metadata. This is done
    automatically when using BlueAPI

    For longer term plans to test plans with BlueAPI in CI,
    see https://github.com/DiamondLightSource/blueapi/issues/1206"""

    def __init__(self, tmp_path: Path) -> None:
        self.path = tmp_path
        self._docs: list[RunStart] = []

    def run_start(self, name: str, start_document: RunStart) -> None:
        if name == "start":
            self._docs.append(start_document)

    def __call__(self, device_name: str | None = None) -> PathInfo:
        template = self._docs[-1].get("detector_file_template")
        if not template:
            raise ValueError("detector_file_template must be set in metadata")
        sub_path = template.format_map(self._docs[-1] | {"device_name": device_name})
        return PathInfo(directory_path=self.path, filename=sub_path)


# See https://github.com/DiamondLightSource/dodal/issues/1455
@pytest.fixture
def jungfrau(tmp_path: Path, run_engine: RunEngine) -> CommissioningJungfrau:
    with init_devices(mock=True):
        base_provider = _BasePathProvider(tmp_path)
        path = AutoMaxIncrementingPathProvider(base_provider)
        detector = CommissioningJungfrau("", "", path)
    set_mock_value(detector._writer.writer_ready, 1)
    run_engine.subscribe(base_provider.run_start, "start")

    return detector


@pytest.fixture
async def beamstop_check_devices(
    aperture_scatterguard,
    attenuator,
    backlight,
    baton,
    detector_motion,
    ipin,
    zebra_shutter,
    xbpm_feedback,
    sim_run_engine,
    run_engine,
):
    async def noop(_):
        await asyncio.sleep(0)

    run_engine.register_command("sleep", noop)
    try:
        async with init_devices(mock=True):
            beamstop = Beamstop("", MagicMock())

        devices = BeamstopCheckDevices(
            aperture_scatterguard=aperture_scatterguard,
            attenuator=attenuator,
            backlight=backlight,
            baton=baton,
            beamstop=beamstop,
            detector_motion=detector_motion,
            ipin=ipin,
            sample_shutter=zebra_shutter,
            xbpm_feedback=xbpm_feedback,
        )
        sim_run_engine.add_read_handler_for(
            devices.sample_shutter, ZebraShutterState.CLOSE
        )
        sim_run_engine.add_handler(
            "locate",
            lambda msg: {"readback": ShutterState.CLOSED},
            "detector_motion-shutter",
        )
        sim_run_engine.add_read_handler_for(ipin.pin_readback, 0.1)

        return devices
    finally:
        run_engine.register_command("sleep", run_engine._sleep)


@pytest.fixture
async def ipin():
    yield i03.ipin.build(connect_immediately=True, mock=True)



from __future__ import annotations

import dataclasses
from collections.abc import Sequence

import numpy as np
from bluesky.callbacks import CallbackBase
from event_model import RunStart


class XRayCentreEventHandler(CallbackBase):
    def __init__(self):
        super().__init__()
        self.xray_centre_results: Sequence[XRayCentreResult] | None = None

    def start(self, doc: RunStart) -> RunStart | None:
        if "xray_centre_results" in doc:
            self.xray_centre_results = [
                XRayCentreResult(**result_dict)
                for result_dict in doc["xray_centre_results"]  # type: ignore
            ]
        return doc


@dataclasses.dataclass
class XRayCentreResult:
    """
    Represents information about a hit from an X-ray centring.

    Attributes:
        centre_of_mass_mm: coordinates in mm of the centre of mass
        bounding_box_mm: coordinates in mm of opposite corners of the bounding box
            containing the crystal
        max_count: The maximum spot count encountered in any one grid box in the crystal
        total_count: The total count across all boxes in the crystal.
        sample_id: The sample id associated with the centre.
    """

    centre_of_mass_mm: np.ndarray
    bounding_box_mm: tuple[np.ndarray, np.ndarray]
    max_count: int
    total_count: int
    sample_id: int | None

    def __eq__(self, o):
        return (
            isinstance(o, XRayCentreResult)
            and o.max_count == self.max_count
            and o.total_count == self.total_count
            and o.sample_id == self.sample_id
            and all(o.centre_of_mass_mm == self.centre_of_mass_mm)
            and all(o.bounding_box_mm[0] == self.bounding_box_mm[0])
            and all(o.bounding_box_mm[1] == self.bounding_box_mm[1])
        )



import numpy as np
from bluesky import plan_stubs as bps
from bluesky.utils import FailedStatus
from dodal.devices.motors import XYZOmegaStage
from ophyd_async.epics.motor import MotorLimitsError

from mx_bluesky.common.utils.exceptions import SampleError


def move_gonio_warn_on_out_of_range(
    gonio: XYZOmegaStage,
    position: np.ndarray | list[float] | tuple[float, float, float],
):
    """
    Throws a SampleException if the specified position is out of range for the
    gonio. Otherwise moves to that position. The check is from ophyd-async
    """
    try:
        yield from bps.mv(
            gonio.x, position[0], gonio.y, position[1], gonio.z, position[2]
        )
    except FailedStatus as fs:
        if isinstance(fs.__cause__, MotorLimitsError):
            raise SampleError(
                "Pin tip centring failed - pin too long/short/bent and out of range"
            ) from fs.__cause__
        else:
            raise fs



import datetime
from pathlib import Path

from bluesky.utils import MsgGenerator
from dodal.common import inject
from dodal.devices.motors import XYZOmegaStage
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.robot import BartRobot

from mx_bluesky.beamlines.aithre_lasershaping.experiment_plans.robot_load_plan import (
    RobotLoadComposite,
    robot_load_and_snapshots_plan,
    robot_unload_plan,
)
from mx_bluesky.beamlines.aithre_lasershaping.parameters.constants import CONST
from mx_bluesky.beamlines.aithre_lasershaping.parameters.robot_load_parameters import (
    AithreRobotLoad,
)


def robot_load_and_snapshot(
    robot: BartRobot = inject("robot"),
    gonio: XYZOmegaStage = inject("gonio"),
    oav: OAV = inject("oav"),
    ptd: PinTipDetection = inject("ptd"),
    tip_offset_microns: float = 0,
    oav_config_file: str = CONST.OAV_CENTRING_FILE,
    sample_puck: int = 0,
    sample_pin: int = 0,
    sample_id: int = 0,
    sample_name: str = "test",
    visit: str = "cm40645-5",
) -> MsgGenerator:
    time_now = datetime.datetime.now()
    year_now = str(time_now.year)
    snapshot_directory = f"/dls/i23/data/{year_now}/{visit}/{sample_name}/snapshots"
    composite = RobotLoadComposite(robot, gonio, oav, gonio)
    params = AithreRobotLoad(
        sample_id=sample_id,
        sample_puck=sample_puck,
        sample_pin=sample_pin,
        snapshot_directory=Path(snapshot_directory),
        visit=visit,
        beamline="BL23I",
    )

    yield from robot_load_and_snapshots_plan(
        composite, params, ptd, tip_offset_microns, oav_config_file
    )


def robot_unload(
    robot: BartRobot = inject("robot"),
    gonio: XYZOmegaStage = inject("gonio"),
    oav: OAV = inject("oav"),
    sample_puck: int = 0,
    sample_pin: int = 0,
    sample_id: int = 0,
    sample_name: str = "test",
    visit: str = "cm40645-5",
) -> MsgGenerator:
    time_now = datetime.datetime.now()
    year_now = str(time_now.year)
    snapshot_directory = f"/dls/i23/data/{year_now}/{visit}/{sample_name}/snapshots"
    composite = RobotLoadComposite(robot, gonio, oav, gonio)
    params = AithreRobotLoad(
        sample_id=sample_id,
        sample_puck=sample_puck,
        sample_pin=sample_pin,
        snapshot_directory=Path(snapshot_directory),
        visit=visit,
        beamline="BL23I",
    )
    yield from robot_unload_plan(composite, params)



from dodal.common.beamlines.beamline_parameters import BEAMLINE_PARAMETER_PATHS

BEAMLINE_PARAMETER_PATHS["i03"] = "tests/test_data/test_beamline_parameters.txt"
BEAMLINE_PARAMETER_PATHS["test"] = "tests/test_data/test_beamline_parameters.txt"
from dodal.beamlines.i03 import *  # type: ignore  # noqa: E402, F403
